import unittest
import scmkl
import anndata as ad
import numpy as np
from scipy.sparse import load_npz


def read_data(mod = 'RNA'):
    """
    Simple function to load example data to run tests.
    """
    x = load_npz(f'../example/data/_MCF7_{mod}_X.npz')
    grouping = np.load(f'../example/data/_{mod}_hallmark_groupings.pkl',
                            allow_pickle = True)
    features = np.load(f'../example/data/_MCF7_{mod}_feature_names.npy', 
                    allow_pickle = True)
    labels = np.load('../example/data/_MCF7_cell_labels.npy', 
                    allow_pickle = True)
    
    return x, grouping, features, labels


class TestCreateAdata(unittest.TestCase):
    """
    This unittest class is designed to test the scmkl.create_adata() 
    function. It creates an anndata.AnnData object and checks the 
    attributes required for scmkl to run.
    """

    def test_create_adata(self):
        """
        This function creates a scmkl AnnData object and checks the 
        train/test split, grouping dictionary, number of dimensions, 
        kernel function, and distance metric.
        """
        # Read-in data
        x, grouping, features, labels = read_data()
        train = ['train']*800
        test = ['test']*200
        train_test = np.array(train + test)
        d = scmkl.calculate_d(len(labels))

        # Creating adata to test
        adata = scmkl.create_adata(X = x, feature_names = features, 
                                   cell_labels = labels, D = d,
                                   group_dict = grouping,
                                   split_data = train_test)

        # Ensuring group dict is intact after object creation
        err_str = ("Genes present in 'adata' group_dict "
                           "not in original grouping.")
        for group in adata.uns['group_dict'].keys():
            for gene in adata.uns['group_dict'][group]:
                
                self.assertIn(gene, grouping[group], err_str)
        
        # Checking that the number of dimensions for n = 1000 is correct
        self.assertEqual(adata.uns['D'], 100, "Incorrect optimal D calculated")

        # Checking default kernel function
        self.assertEqual(adata.uns['kernel_type'].lower(), 'gaussian', 
                         'Default kernel function should be gaussian')

        # Checking default distance metric
        self.assertEqual(adata.uns['distance_metric'], 'euclidean',
                         "Default distance metric should be euclidean")

        # Ensuring train test split is conserved when provided
        train_idx = np.where('train' == train_test)[0]
        train_bool = np.array_equal(adata.uns['train_indices'], train_idx)
        test_idx = np.where('test' == train_test)[0]
        test_bool = np.array_equal(adata.uns['test_indices'], test_idx)
        
        self.assertTrue(train_bool, "Train indices incorrect")
        self.assertTrue(test_bool, "Test indices incorrect")


    def test_obs_retention(self):
        """
        This function is to ensure that `scmkl.create_adata()` is 
        retaining the correct cell labels despite resorting.

        This entails:
            - Making sure cells are sorted as all training, then all 
            testing.
            - Cell names still correspond to the correct row.
            - The training and test indices have been changed to 
            reflect the newly sorted `adata`. 
        """
        # Creating 10x10 matrix where every element in each row is the rows 
        # index (ex. row 5 is 10 5s)
        x = np.array([
            [i for j in range(10)]
            for i in range(10)
        ])
        obs = np.array([f'cell_{i}' for i in range(10)])
        var = np.array([f'feature_{i}' for i in range(10)])
        tt = ['train', 'test']
        train_test = np.array([tt[i%2] for i in range(10)])
        grouping = {'group1' : ['feature_1, feature_2'],
                    'group2' : ['feature_3', 'feature_4']}

        adata = scmkl.create_adata(x, var, train_test, grouping, obs, 
                                   split_data=train_test, 
                                   remove_features=False)
        
        # Making sure order matches obs_names
        # Should be all evens then all odds
        exp_order = ([i for i in range(10) if i%2 == 0] 
                  + [i for i in range(10) if i%2 == 1])
        exp_obs = [f'cell_{i}' for i in exp_order]

        exp_train = np.arange(0, 5)
        exp_test = np.arange(5, 10)
        elements_1d = [adata.X[i][0] for i in range(10)]

        eq_obs = np.all(adata.obs_names == exp_obs)
        eq_train = np.all(adata.uns['train_indices'] == exp_train)
        eq_test = np.all(adata.uns['test_indices'] == exp_test)
        matched_label = np.all(exp_order == elements_1d)

        self.assertTrue(eq_obs, "`.obs_names` are not in the correct order.")
        self.assertTrue(eq_train, "Train indices are not as expected.")
        self.assertTrue(eq_test, "Test indices are not as expected.")
        self.assertTrue(matched_label, 
                        "Elements in `adata.X` no longer match `.obs_names`")


    def test_format_adata(self):
        """
        This function ensures that format adata is working correctly 
        by comparing it's output to `create_adata()`.
        """
        # Read-in data
        x, grouping, features, labels = read_data('RNA')
        train = ['train'] * 800
        test = ['test'] * 200
        train_test = np.array(train + test)
        obs = np.array([f'cell_{i}' for i in range(x.shape[0])])

        # Creating adata to test, remove features must be `False` as it 
        # unreliably resorts features
        adata = scmkl.create_adata(X=x, feature_names=features, 
                                   obs_names=obs,
                                   cell_labels=labels, 
                                   group_dict=grouping,
                                   remove_features=False, 
                                   split_data=train_test)
        
        new_adata = ad.AnnData(X=x)
        new_adata.var_names = features
        new_adata.obs_names = obs
        new_adata.obs['treatment'] = labels
        new_adata = scmkl.format_adata(new_adata, 'treatment', grouping,
                                       remove_features=False, split_data=train_test)

        # Checking that adatas are the same
        expected_keys = {'group_dict', 'seed_obj', 'scale_data', 'D', 
                         'kernel_type', 'distance_metric', 'reduction', 
                         'tfidf', 'labeled_test', 'train_indices', 
                         'test_indices'}
        
        for key in expected_keys:
            self.assertIn(key, new_adata.uns.keys(), 
                          f"{key} is not in `new_adata`.")
            
        eq_mats = np.all(adata.X.toarray() == new_adata.X.toarray())
        eq_var_names = np.all(adata.var_names == new_adata.var_names)
        eq_obs_names = np.all(adata.obs_names == new_adata.obs_names)
        eq_labs = np.all(adata.obs['labels'] == new_adata.obs['labels'])

        self.assertTrue(eq_mats, 
                         "`.X` does not match between adatas.")
        self.assertTrue(eq_var_names,
                         "`.var_names` do not match between adatas.")
        self.assertTrue(eq_obs_names, 
                        "`.obs_names` do not match between adatas.")
        self.assertTrue(eq_labs, 
                        "`.obs['labels']` do not match between adatas.")
        self.assertEqual(adata.uns['D'], new_adata.uns['D'],
                         "`.uns['D']` does not match between adatas.")
        self.assertEqual(adata.uns['kernel_type'], 
                         new_adata.uns['kernel_type'],
                         ("`.uns['kernel_type']` does not match between "
                          "adatas."))
        self.assertEqual(adata.uns['distance_metric'], 
                         new_adata.uns['distance_metric'],
                         ("`.uns['distance_metric']` does not match between "
                          "adatas."))

if __name__ == '__main__':
    unittest.main()


import unittest
import scmkl
import numpy as np
import anndata as ad
from test_calculate_z import create_test_adata


def read_h5ad(mod: str):
    """
    Reads in a h5ad file in ../example/data directory.
    """
    h5_fp = f'../example/data/pbmc_{mod.lower()}.h5ad'
    group_fp = f'../example/data/_{mod.upper()}_azimuth_pbmc_groupings.pkl'
    adata = ad.read_h5ad(h5_fp)

    split_data = np.array(['train']*adata.shape[0])
    split_data[adata.obs['batch'] == '3k'] = 'test'
    transform_data = 'rna' == mod.lower()

    adata = scmkl.format_adata(adata, 'celltypes', group_fp, 
                               split_data=split_data, 
                               allow_multiclass=True, 
                               class_threshold=2000,
                               transform_data=transform_data)

    return adata


class TestOptimizeAlpha(unittest.TestCase):
    """
    This unittest class is used to ensure that scmkl.optimize_alpha() 
    works properly for both unimodal and multimodal runs.
    """
    def test_unimodal_optimize_alpha(self):
        """
        This function ensure that scmkl.optimize_alpha works correctly 
        for unimodal test cases by checking that the output is what we 
        expect.
        """
        adata = create_test_adata()
        alpha_list = np.array([0.01, 0.05, 0.1])

        # Finding optimal alpha
        alpha_star = scmkl.optimize_alpha(adata, 
                                          alpha_array=alpha_list, 
                                          batch_size=60)

        # Checking that optimal alpha is what we expect
        self.assertEqual(alpha_star, 0.1, 
                         ("scmkl.optimize_alpha chose the wrong alpha "
                          "as optimal for unimodal"))
        

    def test_multimodal_optimize_alpha(self):
        """
        This function ensure that scmkl.optimize_alpha works correctly 
        for multimodal test cases by checking that the output is what 
        we expect.
        """
        adatas = [create_test_adata('RNA'), create_test_adata('ATAC')]
        alpha_list = np.array([0.01, 0.05, 0.1])

        alpha_star = scmkl.optimize_alpha(adatas, alpha_array=alpha_list,
                                          batch_size=60)

        self.assertEqual(alpha_star, 0.1, 
                         ("scmkl.optimize_alpha chose the wrong alpha "
                          "as optimal for multimodal run"))

    
    def test_multiclass_optimize_alpha(self):
        """
        This function ensures finding the optimal alpha for best 
        performance via cross validation for multiclass runs is working 
        properly.
        """
        adata = read_h5ad('RNA')
        alphas = np.array([0.1, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8])

        train_dict = scmkl.get_class_train(adata.uns['train_indices'],
                                           adata.obs['labels'], 
                                           adata.uns['seed_obj'])

        alpha_dict = scmkl.optimize_alpha(adata, 
                                          alpha_array=alphas, 
                                          batch_size=26, 
                                          train_dict=train_dict,
                                          metric='F1-Score')

        expected_alphas = {
            'B': 0.1, 
            'CD14+ Monocytes': 0.1, 
            'CD16+ Monocytes': 0.1, 
            'CD4 T': 0.1, 
            'CD8 T': 0.1, 
            'Dendritic': 0.1, 
            'NK': 0.1}

        self.assertDictEqual(alpha_dict, expected_alphas, 
                             ("scmkl.optimize_alpha() for multiclass "
                              "runs returned incorrect values."))
        

    def test_multiview_multiclass_optimize_alpha(self):
        """
        This function ensures that in a multiview and multclass 
        optimize alpha task, everything runs correctly.
        """
        adatas = [read_h5ad('RNA'), read_h5ad('ATAC')]
        alphas = np.array([0.05, 0.1, 0.3, 0.5, 0.6, 0.7, 0.8])

        train_dict = scmkl.get_class_train(adatas[0].uns['train_indices'],
                                           adatas[0].obs['labels'], 
                                           adatas[0].uns['seed_obj'])

        alpha_dict = scmkl.optimize_alpha(adatas, 
                                           alpha_array=alphas, 
                                           train_dict=train_dict,
                                           metric='F1-Score',
                                           batch_size=26)

        expected_alphas = {
            'B': 0.3, 
            'CD14+ Monocytes': 0.3, 
            'CD16+ Monocytes': 0.3, 
            'CD4 T': 0.1, 
            'CD8 T': 0.1, 
            'Dendritic': 0.05, 
            'NK': 0.1}
        
        self.assertDictEqual(alpha_dict, expected_alphas, 
                             ("scmkl.optimize_alpha() for multiclass "
                              "runs returned incorrect values."))
        
    def test_folds_and_processing(self):
        """
        Much of this code adapted from optimize alpha code.
        """
        # Need a matrix to track fold generation with
        x = np.arange(500).reshape(500,1)
        x = np.concatenate([x, x, x], axis=1)
        vars = {f'gene_{i}' for i in range(x.shape[1])}
        labs = ['class_1', 'class_2', 'class_2', 'class_1', 'class_1']*100

        group_dict = {
            'Group1': ['gene_0', 'gene_2'],
            'Group2': ['gene_0', 'gene_1']
        }

        train_test = np.array(300*['train'] + 200*['test'])

        adata = scmkl.create_adata(x, vars, labs, 
                                    group_dict, 
                                    scale_data=False,
                                    split_data=train_test)

        adata = [adata]

        k = 4

        # Only want folds for training samples
        train_indices = adata[0].uns['train_indices'].copy()
        cv_adata = [adata[i][train_indices, :].copy()
                    for i in range(len(adata))]

        folds = scmkl.get_folds(adata[0], k)

        for fold in range(k):

            fold_train, fold_test = folds[fold]
            fold_adata = cv_adata.copy()

            # Downstream functions expect train then test samples in adata(s)
            sort_idcs, fold_train, fold_test = scmkl.sort_samples(fold_train, fold_test)
            fold_adata = scmkl.prepare_fold(fold_adata, sort_idcs, 
                                            fold_train, fold_test)
                
            names = ['Adata ' + str(i + 1) for i in range(len(cv_adata))]

            # Adatas need combined if applicable and kernels calculated 
            fold_adata = scmkl.process_fold(fold_adata, names, tfidf=[False], combination='concatenate', 
                                            batches=10, batch_size=10)

            # Each sample contains only values cooresponding to orig idx
            exp_order = np.concatenate(folds[fold])
            exp_mat = np.repeat(exp_order.reshape(cv_adata[0].shape[0], 1), 
                                cv_adata[0].shape[1], 
                                axis=1)

            # Repeating indices should equal mat (see x initialization)
            sort_err = ("After processing, matrix no longer "
                        "cooresponds with expected indices")
            self.assertTrue(np.array_equal(fold_adata.X, exp_mat), sort_err)

if __name__ == '__main__':
    unittest.main()



import scmkl
import unittest
import numpy as np


class TestGetGeneGrouping(unittest.TestCase):
    """
    This unittest class is used to ensure that 
    scmkl.get_gene_groupings() works properly.
    """

    def test_find_candidates(self):
        """
        This function runs and test the scmkl.find_candidates() 
        function.
        """
        candidates = scmkl.find_candidates('human', key_terms=[' b ', ' t '])

        self.assertEqual(candidates.shape[0], 12, 
                         "Incorrect number of rows, should be 12.")
        self.assertEqual(candidates.shape[1], 3, 
                         "Incorrect number of cols, should be 3.")
        
        no_matching = {39, 55, 0, 161, 115, 100, 2, 1, 10}
        res_matching = set(candidates['No. Key Terms Matching'])

        self.assertSetEqual(no_matching, res_matching, 
                            "Incorrect values in Key term matching column.")


    def test_get_gene_groupings(self):
        """
        This function runs and test the scmkl.get_gene_groupings() 
        function.
        """
        features = np.load('../example/data/_MCF7_RNA_feature_names.npy', 
                            allow_pickle=True)
        
        group_dict = scmkl.get_gene_groupings('CellMarker_2024', 'human',
                                                key_terms=[' b ', ' t '],
                                                min_overlap=2, 
                                                genes=features)
        self.assertEqual(len(group_dict), 161,
                         "Incorrect number of groupings kept w/o blacklist.")
        
        group_dict = scmkl.get_gene_groupings('CellMarker_2024', 'human',
                                                key_terms=[' b ', ' t '],
                                                min_overlap=2, 
                                                blacklist=['blood', 'stomach'],
                                                genes=features)
        self.assertEqual(len(group_dict), 99,
                         "Incorrect number of groupings kept w/ blacklist.")

    
if __name__ == '__main__':
    unittest.main()


import scmkl
import unittest
import numpy as np

from test_calculate_z import create_test_adata


def get_results(opt_alpha=False):
    """
    Runs scmkl from start to finish to test dataframe functions.
    """
    alpha_array = np.array([0.01, 0.1])

    adata = create_test_adata()
    adata = scmkl.calculate_z(adata, batch_size=80)
    results = scmkl.run(adata, alpha_array)

    if opt_alpha:
        results['Alpha_star'] = 0.1

    return results


class TestDataframes(unittest.TestCase):
    """
    This unittest class is used to ensure that 
    the dataframes module is working properly.
    """
    def test_parse_metrics(self):
        """
        This function ensures that scmkl.parse_metrics() works 
        as expected.
        """
        results = get_results(opt_alpha=True)
        metrics = scmkl.parse_metrics(results, include_as=True)

        # Ensuring there are only five rows with Alpha Star == True
        num_opt = np.sum(metrics['Alpha Star'])
        self.assertEqual(num_opt, 5, "Wrong number of alpha_stars.")
        
        # Ensuring only the correct alphas are present
        self.assertSetEqual({0.1, 0.01}, set(metrics['Alpha']),
                            "Alphas should be 0.1 and 0.01.")
        

    def test_parse_weights(self):
        """
        This function ensures that scmkl.parse_weights() works 
        as expected.
        """
        results = get_results(opt_alpha=True)
        weights = scmkl.parse_weights(results, include_as=True)
        
        # Ensuring there are only five rows with Alpha Star == True
        num_opt = np.sum(weights['Alpha Star'])
        self.assertEqual(num_opt, 50, "Wrong number of alpha_stars.")
        
        # Ensuring only the correct alphas are present
        self.assertSetEqual({0.1, 0.01}, set(weights['Alpha']),
                            "Alphas should be 0.1 and 0.01.")


    def test_get_metrics(self):
        """
        This function ensures that scmkl.get_metrics() works 
        as expected.
        """
        # Creating results to run in normal mode
        results = get_results(opt_alpha=True)
        metrics = scmkl.get_metrics(results, include_as=True)

        # Ensuring there are only five rows with Alpha Star == True
        num_opt = np.sum(metrics['Alpha Star'])
        self.assertEqual(num_opt, 5, "Wrong number of alpha_stars.")
        
        # Ensuring only the correct alphas are present
        self.assertSetEqual({0.1, 0.01}, set(metrics['Alpha']),
                            "Alphas should be 0.1 and 0.01.")

        # Creating results to run with rfiles
        rfiles = {'rep_1' : results, 'rep_2' : results}
        metrics = scmkl.get_metrics(rfiles, include_as=True)
        
        # Ensuring there are only five rows with Alpha Star == True
        num_opt = np.sum(metrics['Alpha Star'])
        self.assertEqual(num_opt, 10, "Wrong number of alpha_stars.")
        
        # Ensuring only the correct alphas are present
        self.assertSetEqual({0.1, 0.01}, set(metrics['Alpha']),
                            "Alphas should be 0.1 and 0.01.")


    def test_get_weights(self):
        """
        This function ensures that scmkl.get_weights() works 
        as expected.
        """
        # Creating results to run in normal mode
        results = get_results(opt_alpha=True)
        weights = scmkl.get_weights(results, include_as=True)

        # Ensuring there are only five rows with Alpha Star == True
        num_opt = np.sum(weights['Alpha Star'])
        self.assertEqual(num_opt, 50, "Wrong number of alpha_stars.")
        
        # Ensuring only the correct alphas are present
        self.assertSetEqual({0.1, 0.01}, set(weights['Alpha']),
                            "Alphas should be 0.1 and 0.01.")

        # Creating results to run with rfiles
        rfiles = {'rep_1' : results, 'rep_2' : results}
        weights = scmkl.get_weights(rfiles, include_as=True)
        
        # Ensuring there are only five rows with Alpha Star == True
        num_opt = np.sum(weights['Alpha Star'])
        self.assertEqual(num_opt, 100, "Wrong number of alpha_stars.")
        
        # Ensuring only the correct alphas are present
        self.assertSetEqual({0.1, 0.01}, set(weights['Alpha']),
                            "Alphas should be 0.1 and 0.01.")
        

    def test_get_selection(self):
        """
        This function ensures that scmkl.get_selection() works 
        as expected.
        """
        # Creating results to run in normal mode
        results = get_results(opt_alpha=True)
        
        # Creating results to run with rfiles
        rfiles = {'rep_1' : results, 'rep_2' : results}
        weights = scmkl.get_weights(rfiles, include_as=True)
        
        selection = scmkl.get_selection(weights, order_groups=True)
        
        # Checking that each selection < number of reps
        cor_reps = selection['Selection'] <= len(rfiles)
        self.assertTrue(np.all(cor_reps), "Selection is greater than expect.")

        # Ensuring all groups are present per alpha
        num_rows = len(rfiles)*50
        self.assertEqual(num_rows, len(selection), 
                         "Incorrect number of rows in selection df.")
        
        # Checking multiclass output
        results = {'B' : results, 'T' : results, 'Mono' : results, 
                   'Classes' : ['B', 'T', 'Mono']}
        
        weights = scmkl.get_weights(results, include_as=False)
        selection = scmkl.get_selection(weights, False)
        
        self.assertIn('Class', selection.columns,
                      "`'Class'` column missing from df.")
        self.assertEqual(selection.shape[0], 300, 
                         "df is wrong len for 2 alphas, 50 groups, 3 classes.")


    def test_read_gtf(self):
        """
        Checks that scmkl.read_gtf() is functioning properly.
        """
        # Reading in gtf file
        gtf_fp = '../example/data/_hg38_subset_protein_coding.annotation.gtf'
        gtf = scmkl.read_gtf(gtf_fp, filter_to_coding=True)

        # Checking columns
        gtf_cols = set(gtf.columns)
        expected_cols = {'frame', 'attribute', 'end', 'strand', 'feature', 
                         'gene_name', 'start', 'source', 'score', 'chr'}
        self.assertSetEqual(gtf_cols, expected_cols, 
                            "Incorrect columns in gtf dataframe.")

        # Ensuring only protein coding genes are present
        is_protein_coding = ['protein_coding' in attr 
                             for attr in gtf['attribute']]
        self.assertEqual(np.sum(is_protein_coding), len(gtf),
                         "Nonprotein coding annotations present.")


if __name__ == '__main__':
    unittest.main()


import unittest
import scmkl
import numpy as np
from test_train_model import train_scmkl


class TestTest(unittest.TestCase):
    """
    This unittest class is used to evaluate whether the functions in 
    scmkl.test are working properly.
    """

    def test_predict(self):
        """
        This function ensures that the output model from 
        scmkl.predict() returns as expected by checking the number of 
        coefficients.
        """
        # Creating adata
        adata = train_scmkl()

        metrics = ['AUROC', 'Accuracy', 'F1-Score', 'Precision', 'Recall']

        predictions, metrics_dict = scmkl.predict(adata, metrics)

        # Checking that metrics are present
        mets_check = [met in metrics_dict.keys() 
                      for met in metrics]
        
        self.assertTrue(np.all(mets_check), 
                        "There are missing metrics in metrics_dict")
        
        # Checking that output is binary
        self.assertTrue(np.unique(predictions).shape[0] == 2, 
                        "There are not only two classes in predictions")


    def test_find_selected_groups(self):
        """
        This function tests the scmkl.find_selected_groups() function 
        by ensuring the expected number of groups are output by the 
        function for the trained model.
        """
        # Training scMKL
        adata = train_scmkl()

        # Finding the selected groups from the model
        selected_groups = scmkl.find_selected_groups(adata)

        # Ensuring the number of output groups are expected
        self.assertEqual(len(selected_groups), 44,
                         "Incorrect number of selected groups")


if __name__ == '__main__':
    unittest.main()


import anndata as ad
import numpy as np
import time
import tracemalloc

from scmkl.train_model import train_model
from scmkl.test import predict, find_selected_groups


def run(adata: ad.AnnData, alpha_list: np.ndarray, 
        metrics: list | None = None, 
        return_probs: bool=False) -> dict:
    """
    Wrapper function for training and test with multiple alpha values.
    Returns metrics, predictions, group weights, and resource usage.

    Parameters
    ----------
    adata : ad.AnnData 
        A processed `ad.AnnData` with `'Z_train'`, `'Z_test'`, and 
        `'group_dict'` keys in `adata.uns`.
    
    alpha_list : np.ndarray 
        Sparsity values to create models with. Alpha refers to the 
        penalty parameter in Group Lasso. Larger alphas force group 
        weights to shrink towards zero while smaller alphas apply a 
        lesser penalty to kernal weights. Values too large will results 
        in models that weight all groups as zero.

    metrics : list[str]
        Metrics that should be calculated on predictions. Options are 
        `['AUROC', 'F1-Score', 'Accuracy', 'Precision', 'Recall']`. 
        When set to `None`, all metrics are calculated.
    
    Returns
    -------
    results : dict
        Results with keys and values: 

        `'Metrics'` (dict): 
        A nested dictionary as `[alpha][metric] = value`.

        `'Group_names'` (np.ndarray): 
        Array of group names used in model(s).
    
        `'Selected_groups'` (dict): 
        A nested dictionary as `[alpha] = np.array([nonzero_groups])`.
        Nonzero groups are groups that had a kernel weight above zero.

        `'Norms'` (dict): 
        A nested dictionary as `[alpha] = np.array([kernel_weights])`
        Order of `kernel_weights` is respective to `'Group_names'` 
        values.

        `'Observed'` (np.nparray): 
        An array of ground truth cell labels from the test set.

        `'Predictions'` (dict): 
        A nested dictionary as `[alpha] = predicted_class` respective 
        to `'Observations'` for `alpha`.

        `'Test_indices'` (np.array: 
        Indices of samples respective to adata used in the training 
        set.

        `'Model'` (dict): 
        A nested dictionary where `[alpha] = celer.GroupLasso` object 
        for `alpha`.

        `'RAM_usage'` (dict): 
        A nested dictionary with memory usage in GB after 
        training models for each `alpha`.

    Examples
    --------
    >>> results = scmkl.run(adata = adata, 
    ...                     alpha_list = np.array([0.05, 0.1, 0.5]))
    >>> results
    dict_keys(['Metrics', 'Selected_groups', 'Norms', 'Predictions', 
    ...        'Observed', 'Test_indices', 'Group_names', 'Models', 
    ...        'Train_time', 'RAM_usage'])
    >>>
    >>> alpha values
    >>> results['Metrics'].keys()
    dict_keys([0.05, 0.1, 0.5])
    >>>
    >>> results['Metrics'][0.05]
    {'AUROC': 0.9859,
    'Accuracy': 0.945,
    'F1-Score': 0.9452736318407959,
    'Precision': 0.9405940594059405,
    'Recall': 0.95}
    """
    if metrics is None:
        metrics = ['AUROC', 'F1-Score','Accuracy', 'Precision', 'Recall']

    # Initializing variables to capture metrics
    group_names = list(adata.uns['group_dict'].keys())
    preds = {}
    group_norms = {}
    mets_dict = {}
    selected_groups = {}
    train_time = {}
    models = {}
    probs = {}

    D = adata.uns['D']

    # Generating models for each alpha and outputs
    tracemalloc.start()
    for alpha in alpha_list:
        
        print(f'  Evaluating model. Alpha: {alpha}', flush = True)

        train_start = time.time()

        adata = train_model(adata, group_size= 2*D, alpha = alpha)

        if return_probs:
            alpha_res = predict(adata, 
                                metrics = metrics,
                                return_probs = return_probs)
            preds[alpha], mets_dict[alpha], probs[alpha] = alpha_res

        else:
            alpha_res = predict(adata, 
                                metrics = metrics,
                                return_probs = return_probs)
            preds[alpha], mets_dict[alpha] = alpha_res

        selected_groups[alpha] = find_selected_groups(adata)

        kernel_weights = adata.uns['model'].coef_
        group_norms[alpha] = [
            np.linalg.norm(kernel_weights[i * 2 * D : (i + 1) * 2 * D - 1])
            for i in np.arange(len(group_names))
            ]
        
        models[alpha] = adata.uns['model']
        
        train_end = time.time()
        train_time[alpha] = train_end - train_start

    # Combining results into one object
    results = {}
    results['Metrics'] = mets_dict
    results['Selected_groups'] = selected_groups
    results['Norms'] = group_norms
    results['Predictions'] = preds
    results['Observed'] = adata.obs['labels'].iloc[adata.uns['test_indices']]
    results['Test_indices'] = adata.uns['test_indices']
    results['Group_names']= group_names
    results['Models'] = models
    results['Train_time'] = train_time
    results['RAM_usage'] = f'{tracemalloc.get_traced_memory()[1]/1e9} GB'
    results['Probabilities'] = probs

    return results


import gc
import numpy as np
import anndata as ad
from sklearn.model_selection import StratifiedKFold

from scmkl.tfidf_normalize import tfidf_normalize
from scmkl.calculate_z import calculate_z
from scmkl.train_model import train_model
from scmkl.multimodal_processing import multimodal_processing
from scmkl.test import predict
from scmkl.one_v_rest import get_class_train
from scmkl.create_adata import sort_samples


# Array of alphas to be used if not provided
default_alphas = np.round(np.linspace(1.9, 0.05, 10),2)


def stop_early(metric_array, alpha_idx, fold_idx):
    """
    Assumes smallest alpha comes first.
    """
    # Must be at least two metrics from two alphas to compare
    if alpha_idx <= 0:
        return False
    
    cur_met = metric_array[alpha_idx, fold_idx]
    last_met = metric_array[alpha_idx - 1, fold_idx]

    if cur_met < last_met:
        return True
    else:
        return False


def sort_alphas(alpha_array: np.ndarray):
    """
    Sorts alphas from smallest to largest.
    """
    order = np.argsort(alpha_array)
    alpha_array = alpha_array[order]

    return alpha_array


def get_folds(adata: ad.AnnData, k: int):
    """
    With labels of samples for cross validation and number of folds, 
    returns the indices and label for each k-folds.

    Parameters
    ----------
    adata : ad.AnnData
        `AnnData` object containing `'labels'` column in `.obs` and 
        `'train_indices'` in `.uns`.

    k : int
        The number of folds to perform cross validation over.

    Returns
    -------
    folds : dict
        A dictionary with keys being [0, k) and values being a tuple 
        with first element being training sample indices and the second 
        being testing value indices.

    Examples
    --------
    >>> adata = scmkl.create_adata(...)
    >>> folds = scmkl.get_folds(adata, k=4)
    >>>
    >>> train_fold_0, test_fold_0 = folds[0]
    >>>
    >>> train_fold_0
    array([  0,   3,   5,   6,   8,   9,  10,  11,  12,  13])
    >>> test_fold_0
    array([  1,   2,  4,  7])
    """
    y = adata.obs['labels'][adata.uns['train_indices']].copy()

    # Creating dummy x prevents holding more views in memory
    x = y.copy()

    skf = StratifiedKFold(n_splits=k, shuffle=True, random_state=100)

    folds = dict()
    for fold, (fold_train, fold_test) in enumerate(skf.split(x, y)):
        folds[fold] = (fold_train, fold_test)

    return folds


def prepare_fold(fold_adata, sort_idcs, fold_train, fold_test):
    """
    Reorders adata samples, reassigns each `adata.uns['train_indices' & 
    'test_indices']`, and removes sigmas if present.

    Parameters
    ----------
    fold_adata : list[ad.AnnData]
        A `list` of `AnnData` objects to be reordered based on fold 
        indices.

    sort_idcs: np.ndarray
        The indices that will sort `AnnData`s as all train samples then 
        all test.

    fold_train : np.ndarray
        The indices of training sample for respective fold.

    fold_test : np.ndarray
        The indices of testing samples for respective fold.

    Returns
    -------
    fold_adata : list[ad.AnnData]
        The `list` of `AnnData`s with each `AnnData` sorted as all 
        training samples then all testing samples. 
        `adata.uns['train_indices' & 'test_indices']` are also updated 
        to reflect new sample positions.

    Examples
    --------
    >>> adata = [scmkl.create_adata(...)]
    >>> folds = scmkl.get_folds(adata[0], k=4)
    >>> sort_idcs, fold_train, fold_test = sort_samples(fold_train, 
    ...                                                 fold_test)
    >>> fold_adata = scmkl.prepare_fold(adata, sort_idcs, 
    ...                                 fold_train, fold_test)
    """
    for i in range(len(fold_adata)):
        fold_adata[i] = fold_adata[i][sort_idcs]
        fold_adata[i].uns['train_indices'] = fold_train
        fold_adata[i].uns['test_indices'] = fold_test
    
    # Need to recalculate sigmas for new fold train
    for i in range(len(fold_adata)):
        if 'sigma' in fold_adata[i].uns_keys():
            del fold_adata[i].uns['sigma']

    return fold_adata


def process_fold(fold_adata, names, tfidf, combination, batches, batch_size):
    """
    Combines adata if needed, estimates sigmas, and calculates kernels 
    for model training and evaluation.

    Parameters
    ----------
    fold_adata : list[ad.AnnData]
        A `list` of `AnnData` objects to combine and calculate kernels 
        for.

    names : list
        A list of names respective to each `AnnData` in `fold_adata` 
        for verbose outputs.

    tfidf : list[bool]
        A boolean list indicating whether or not to perform TF-IDF 
        transformation for each adata respective to `fold_adata`.

    combination : str
        The method of combining `AnnData` objects passed to 
        `ad.concatenate()`. Ignored if `len(fold_adata) == 1`.

    batches : int
        The number of batches for kernel width (sigma) estimation.

    batch_size : int
        The number of samples to include in each batch of kernel width 
        (sigma) estimation.

    Examples
    --------
    >>> adata = [scmkl.create_adata(...)]
    >>> tfidf = [False]
    >>> names = ['adata1']
    >>> folds = scmkl.get_folds(adata[0], k=4)
    >>> sort_idcs, fold_train, fold_test = sort_samples(fold_train, 
    ...                                                 fold_test)
    >>> fold_adata = scmkl.prepare_fold(adata, sort_idcs, 
    ...                                 fold_train, fold_test)
    >>> fold_adata = scmkl.process_fold(fold_adata, names, tfidf)
    """
    if 1 < len(fold_adata):
        fold_adata = multimodal_processing(fold_adata, names, tfidf, 
                                           combination, batches, 
                                           batch_size, False)
    else:
        fold_adata = fold_adata[0]
        if tfidf[0]:
            fold_adata = tfidf_normalize(fold_adata, binarize=True)
        fold_adata = calculate_z(fold_adata, n_features= 5000, 
                                 batches=batches, batch_size=batch_size)
        
    return fold_adata


def bin_optimize_alpha(adata: list[ad.AnnData], 
                       group_size: int | None=None, 
                       tfidf: list[bool]=False, 
                       alpha_array: np.ndarray=default_alphas, 
                       k: int=4, metric: str='AUROC', 
                       early_stopping: bool=False,
                       batches: int=10, batch_size: int=100,
                       combination: str='concatenate'):
    """
    Iteratively train a grouplasso model and update alpha to find the 
    parameter yielding best performing sparsity via k-fold cross 
    validation. This function currently only works for binary 
    experiments. Called by `scmkl.optimize_alpha()`.

    Parameters
    ----------
    adata : list[ad.AnnData]
        List of `ad.AnnData`(s) with `'Z_train'` and `'Z_test'` in 
        `adata.uns.keys()`.

    group_size : None | int
        Argument describing how the features are grouped. If `None`, 
        `2 * adata.uns['D']` will be used. For more information see 
        [celer documentation](https://mathurinm.github.io/celer/
        generated/celer.GroupLasso.html).

    tfidf : list | bool
        If `False`, no data will be TF-IDF transformed. If 
        `type(adata) is list` and TF-IDF transformation is desired for 
        all or some of the data, a bool list corresponding to `adata` 
        must be provided. To simply TF-IDF transform `adata` when 
        `type(adata) is ad.AnnData`, use `True`.
    
    alpha_array : np.ndarray
        Array of all alpha values to be tested.

    k : int
        Number of folds to perform cross validation over.
            
    metric : str
        Which metric to use to optimize alpha. Options are `'AUROC'`, 
        `'Accuracy'`, `'F1-Score'`, `'Precision'`, and `'Recall'`.

    batches : int
        The number of batches to use for the distance calculation.
        This will average the result of `batches` distance calculations
        of `batch_size` randomly sampled cells. More batches will converge
        to population distance values at the cost of scalability.

    batch_size : int
        The number of cells to include per batch for distance
        calculations. Higher batch size will converge to population
        distance values at the cost of scalability. If 
        `batches*batch_size > num_training_cells`, `batch_size` will be 
        reduced to `int(num_training_cells/batches)`.

    Returns
    -------
    alpha_star : float
        The best performing alpha value from cross validation on 
        training data.

    Examples
    --------
    >>> alpha_star = scmkl.optimize_alpha(adata)
    >>> alpha_star
    0.1
    """    
    assert isinstance(k, int) and k > 0, "'k' must be positive"

    import warnings 
    warnings.filterwarnings('ignore')

    alpha_array = sort_alphas(alpha_array)

    # Only want folds for training samples
    train_indices = adata[0].uns['train_indices'].copy()
    cv_adata = [adata[i][train_indices, :].copy()
                for i in range(len(adata))]

    folds = get_folds(adata[0], k)

    metric_array = np.zeros((len(alpha_array), k))

    for fold in range(k):

        fold_train, fold_test = folds[fold]
        fold_adata = cv_adata.copy()

        # Downstream functions expect train then test samples in adata(s)
        sort_idcs, fold_train, fold_test = sort_samples(fold_train, fold_test)
        fold_adata = prepare_fold(fold_adata, sort_idcs, 
                                  fold_train, fold_test)
            
        names = ['Adata ' + str(i + 1) for i in range(len(cv_adata))]

        # Adatas need combined if applicable and kernels calculated 
        fold_adata = process_fold(fold_adata, names, tfidf, combination, 
                                  batches, batch_size)

        for i, alpha in enumerate(alpha_array):

            fold_adata = train_model(fold_adata, group_size, alpha=alpha)
            _, metrics = predict(fold_adata, metrics=[metric])
            metric_array[i, fold] = metrics[metric]

            # If metrics are decreasing, cv stopped and moving to next fold
            end_fold = stop_early(metric_array, alpha_idx=i, fold_idx=fold)
            if end_fold and early_stopping:
                break

        del fold_adata
        gc.collect()

    del cv_adata
    gc.collect()

    # Need highest performing alpha for given metric
    alpha_star = alpha_array[np.argmax(np.mean(metric_array, axis = 1))]

    return alpha_star


def multi_optimize_alpha(adata: list[ad.AnnData], group_size: int, 
                         tfidf: list[bool]=[False], 
                         alpha_array: np.ndarray=default_alphas, k: int=4, 
                         metric: str='AUROC', early_stopping: bool=False,
                         batches: int=10, batch_size: int=100, 
                         force_balance: bool=True, combination: str='concatenate',
                         train_dict: dict=None):
    """
    Wrapper function for running k-fold cross validation for every 
    label in a multiclass experiment. Called by 
    `scmkl.optimize_alpha()`. 

    Parameters
    ----------
    adata : list[ad.AnnData]
        List of `ad.AnnData`(s) with `'Z_train'` and `'Z_test'` in 
        `adata.uns.keys()`.

    group_size : None | int
        Argument describing how the features are grouped. If `None`, 
        `2 * adata.uns['D']` will be used. For more information see 
        [celer documentation](https://mathurinm.github.io/celer/
        generated/celer.GroupLasso.html).

    tfidf : list[bool]
        If `False`, no data will be TF-IDF transformed. If 
        `type(adata) is list` and TF-IDF transformation is desired for 
        all or some of the data, a bool list corresponding to `adata` 
        must be provided. To simply TF-IDF transform `adata` when 
        `type(adata) is ad.AnnData`, use `True`.
    
    alpha_array : np.ndarray
        Array of all alpha values to be tested.

    k : int
        Number of folds to perform cross validation over.
            
    metric : str
        Which metric to use to optimize alpha. Options are `'AUROC'`, 
        `'Accuracy'`, `'F1-Score'`, `'Precision'`, and `'Recall'`.

    batches : int
        The number of batches to use for the distance calculation.
        This will average the result of `batches` distance calculations
        of `batch_size` randomly sampled cells. More batches will converge
        to population distance values at the cost of scalability.

    batch_size : int
        The number of cells to include per batch for distance
        calculations. Higher batch size will converge to population
        distance values at the cost of scalability. If 
        `batches*batch_size > num_training_cells`, `batch_size` will be 
        reduced to `int(num_training_cells/batches)`.

    force_balance: bool
        If `True`, training sets will be balanced to reduce class label 
        imbalance for each iteration. Defaults to `False`.

    other_factor : float
        The ratio of cells to sample for the other class for each 
        model. For example, if classifying B cells with 100 B cells in 
        training, if `other_factor=1`, 100 cells that are not B cells 
        will be trained on with the B cells. This will be done for each 
        fold for each class if `force_balance` is `True`. 

    combination: str
        How should multiple views of data be combined. For more details 
        see ad.concat.

    train_dict: dict
        A `dict` where each key is a class label and values are are the 
        indices to be trained with for that class for class balance. 
        All values must be present in each adata.uns['train_indices'].

    Returns
    -------
    alpha_star : dict
        A dictionary with keys being class labels and values being the 
        best performing alpha parameter for that class as a float.
    """
    classes = np.unique(adata[0].obs['labels'])
    orig_labels = adata[0].obs['labels'].to_numpy().copy()
    orig_train = adata[0].uns['train_indices'].copy()

    if train_dict:
        train_idcs = train_dict
    else:
        if force_balance:
            train_idcs = get_class_train(adata[0].uns['train_indices'], 
                                         adata[0].obs['labels'], 
                                         adata[0].uns['seed_obj'])
        else:
            train_idcs = {ct: adata[0].uns['train_indices'].copy()
                          for ct in classes}

    opt_alpha_dict = dict()

    for cl in classes:
        temp_classes = orig_labels.copy()
        temp_classes[temp_classes != cl] = 'other'

        for i in range(len(adata)):
            adata[i].obs['labels'] = temp_classes.copy()
            adata[i].uns['train_indices'] = train_idcs[cl]
        
        opt_alpha_dict[cl] = bin_optimize_alpha(adata, group_size, tfidf, 
                                                alpha_array, k, metric, 
                                                early_stopping, batches, 
                                                batch_size, combination)     
    
    # Global adata obj will be permanently changed if not reset
    for i in range(len(adata)):
        adata[i].obs['labels'] = orig_labels
        adata[i].uns['train_indices'] = orig_train

    return opt_alpha_dict


def optimize_alpha(adata: ad.AnnData | list[ad.AnnData], 
                   group_size: int | None=None, tfidf: None | list[bool]=None, 
                   alpha_array: np.ndarray=default_alphas, k: int=4, 
                   metric: str='AUROC', early_stopping: bool=False,
                   batches: int=10, batch_size: int=100, 
                   combination: str='concatenate', force_balance: bool=True,
                   train_dict: dict=None):
    """
    K-fold cross validation for optimizing alpha hyperparameter using 
    training indices. 

    Parameters
    ----------
    adata : list[ad.AnnData]
        List of `ad.AnnData`(s) with `'Z_train'` and `'Z_test'` in 
        `adata.uns.keys()`.

    group_size : None | int
        Argument describing how the features are grouped. If `None`, 
        `2 * adata.uns['D']` will be used. For more information see 
        [celer documentation](https://mathurinm.github.io/celer/
        generated/celer.GroupLasso.html).

    tfidf : list[bool]
        If `False`, no data will be TF-IDF transformed. If 
        `type(adata) is list` and TF-IDF transformation is desired for 
        all or some of the data, a bool list corresponding to `adata` 
        must be provided. To simply TF-IDF transform `adata` when 
        `type(adata) is ad.AnnData`, use `True`.
    
    alpha_array : np.ndarray
        Array of all alpha values to be tested.

    k : int
        Number of folds to perform cross validation over.
            
    metric : str
        Which metric to use to optimize alpha. Options are `'AUROC'`, 
        `'Accuracy'`, `'F1-Score'`, `'Precision'`, and `'Recall'`.

    batches : int
        The number of batches to use for the distance calculation.
        This will average the result of `batches` distance calculations
        of `batch_size` randomly sampled cells. More batches will converge
        to population distance values at the cost of scalability.

    batch_size : int
        The number of cells to include per batch for distance
        calculations. Higher batch size will converge to population
        distance values at the cost of scalability. If 
        `batches*batch_size > num_training_cells`, `batch_size` will be 
        reduced to `int(num_training_cells/batches)`.

    force_balance: bool
        If `True`, training sets will be balanced to reduce class label 
        imbalance for each iteration. Defaults to `False`.

    other_factor : float
        The ratio of cells to sample for the other class for each 
        model. For example, if classifying B cells with 100 B cells in 
        training, if `other_factor=1`, 100 cells that are not B cells 
        will be trained on with the B cells. This will be done for each 
        fold for each class if `force_balance` is `True`. 

    combination: str
        How should multiple views of data be combined. For more details 
        see ad.concat.

    train_dict: dict
        A `dict` where each key is a class label and values are are the 
        indices to be trained with for that class for class balance. 
        All values must be present in each adata.uns['train_indices'].

    Returns
    -------
    alpha_star : float | dict
        If number of classes is more than 2, a dictionary with keys 
        being class labels and values being the best performing alpha 
        parameter for that class as a float. Else, a float for 
        comparing the two classes.
    """
    # Need singe-view runs to be iterable
    if isinstance(adata, ad.AnnData):
        adata = [adata.copy()]

    if isinstance(tfidf, type(None)):
        tfidf = len(adata)*[False]

    is_multi = 2 < len(set(adata[0].obs['labels']))
    
    if isinstance(group_size, type(None)):
        group_size = 2*adata[0].uns['D']


    if is_multi:
        alpha_star = multi_optimize_alpha(adata, group_size, tfidf, 
                                          alpha_array, k, metric, 
                                          early_stopping, batches, 
                                          batch_size, force_balance, 
                                          combination, train_dict)
        
    else:
        alpha_star = bin_optimize_alpha(adata, group_size, tfidf, 
                                        alpha_array, k, metric, 
                                        early_stopping, batches, 
                                        batch_size, combination)
        
    gc.collect()

    return alpha_star



import numpy as np
import anndata as ad
import scipy
import pandas as pd
import gc
import warnings


def filter_features(feature_names: np.ndarray, group_dict: dict):
    """
    Function to remove features only in feature names or group_dict.
    Any features not included in group_dict will be removed from the
    matrix. Also puts the features in the same relative order (of 
    included features)
    
    Parameters
    ----------
    feature_names : np.ndarray
        Numpy array of corresponding feature names.

    group_dict : dict
        Dictionary containing feature grouping information.
                 Example: {geneset: np.array(gene_1, gene_2, ..., 
                 gene_n)}
    Returns
    -------
    feature_names : np.ndarray
        Numpy array of corresponding feature names from group_dict.

    group_dict : dict
        Dictionary containing features overlapping input grouping 
        information and full feature names.
    """
    group_features = set()
    feature_set = set(feature_names)

    # Store all objects in dictionary in set
    for group in group_dict.keys():
        group_features.update(set(group_dict[group]))

        # Finds intersection between group features and features in data
        # Converts to nd.array and sorts to preserve order of feature names
        group_feats = list(feature_set.intersection(set(group_dict[group])))
        group_dict[group] = np.sort(np.array(group_feats))

    # Only keeping groupings that have at least two features
    group_dict = {group : group_dict[group] for group in group_dict.keys()
                  if len(group_dict[group]) > 1}

    group_features = np.array(list(group_features.intersection(feature_set)))

    return group_features, group_dict


def multi_class_split(y: np.ndarray, seed_obj: np.random._generator.Generator, 
                      class_threshold: str | int | None=None, 
                      train_ratio: float=0.8):
    """
    Function for calculating the training and testing cell positions 
    for multiclass data sets.

    Parameters
    ----------
    y : array_like
        Should be an iterable object cooresponding to samples in 
        `ad.AnnData` object.

    seed_obj : np.random._generator.Generator
        Seed used to randomly sample and split data.

    train_ratio : float
        Ratio of number of training samples to entire data set. 
        Note: if a threshold is applied, the ratio training samples 
        may decrease depending on class balance and `class_threshold`
        parameter.

    class_threshold : str | int
        If is type `int`, classes with more samples than 
        class_threshold will be sampled. If `'median'`, 
        samples will be sampled to the median number of samples per 
        class.

    Returns
    -------
    train_indices : np.ndarray
        Indices for training samples.

    test_indices : np.ndarray
        Indices for testing samples.
    """
    uniq_labels = np.unique(y)

    # Finding indices for each cell class
    class_positions = {class_ : np.where(y == class_)[0] 
                       for class_ in uniq_labels}
    
    # Capturing training indices while maintaining original class proportions
    train_samples = {class_ : seed_obj.choice(class_positions[class_], 
                                              int(len(class_positions[class_])
                                                  * train_ratio), 
                                              replace = False)
                        for class_ in class_positions.keys()}
    
    # Capturing testing indices while maintaining original class proportions
    test_samples = {class_ : np.setdiff1d(class_positions[class_], 
                                          train_samples[class_])
                    for class_ in class_positions.keys()}
    
    # Applying threshold for samples per class
    if class_threshold == 'median':
        cells_per_class = [len(values) for values in train_samples.values()]
        class_threshold = int(np.median(cells_per_class))

    if isinstance(class_threshold, int):
        # Down sample to class_threshold
        for class_ in train_samples.keys():
            if len(train_samples[class_]) > class_threshold:
                train_samples[class_] = seed_obj.choice(train_samples[class_], 
                                                        class_threshold)
            
    train_indices = np.array([idx for class_ in train_samples.keys()
                                  for idx in train_samples[class_]])
    
    test_indices = np.array([idx for class_ in test_samples.keys()
                                 for idx in test_samples[class_]])
    
    return train_indices, test_indices


def binary_split(y: np.ndarray, train_indices: np.ndarray | None=None, 
                  train_ratio: float=0.8,
                  seed_obj: np.random._generator.Generator=np.random.default_rng(100)):
    """
    Function to calculate training and testing indices for given 
    dataset. If train indices are given, it will calculate the test 
    indices. If train_indices == None, then it calculates both indices, 
    preserving the ratio of each label in y

    Parameters
    ----------
    y : np.ndarray
        Numpy array of cell labels. Can have any number of classes 
        for this function.

    train_indices : np.ndarray | None
        Optional array of pre-determined training indices

    train_ratio : float
        Decimal value ratio of features in training/testing sets

    seed_obj : np.random._generator.Generator
        Numpy random state used for random processes. Can be 
        specified for reproducubility or set by default.
    
    
    Returns
    -------
    train_indices : np.ndarray
        Array of indices of training cells.

    test_indices : np.ndarray:
        Array of indices of testing cells.
    """
    # If train indices aren't provided
    if train_indices is None:

        unique_labels = np.unique(y)
        train_indices = []

        for label in unique_labels:

            # Find indices of each unique label
            label_indices = np.where(y == label)[0]

            # Sample these indices according to train ratio
            n = int(len(label_indices) * train_ratio)
            train_label_indices = seed_obj.choice(label_indices, n, 
                                                  replace = False)
            train_indices.extend(train_label_indices)
    else:
        assert len(train_indices) <= len(y), ("More train indices than there "
                                              "are samples")

    train_indices = np.array(train_indices)

    # Test indices are the indices not in the train_indices
    test_indices = np.setdiff1d(np.arange(len(y)), train_indices, 
                                assume_unique = True)

    return train_indices, test_indices


def calculate_d(num_samples : int):
    """
    This function calculates the optimal number of dimensions for 
    performance. See https://doi.org/10.48550/arXiv.1806.09178 for more
    information.

    Parameters
    ----------
    num_samples : int
        The number of samples in the data set including both training
        and testing sets.

    Returns
    -------
    d : int
        The optimal number of dimensions to run scMKL with the given 
        data set.

    Examples
    --------
    >>> raw_counts = scipy.sparse.load_npz('MCF7_counts.npz')
    >>>
    >>> num_cells = raw_counts.shape[0]
    >>> d = scmkl.calculate_d(num_cells)
    >>> d
    161
    """
    d = int(np.sqrt(num_samples)*np.log(np.log(num_samples)))

    return int(np.max([d, 100]))


def sort_samples(train_indices, test_indices):
    """
    Ensures that samples in adata obj are all training, then all 
    testing.

    Parameters
    ----------
    train_indices : np.ndarray
        Indices in ad.AnnData object for training.
    
    test_indices : np.ndarray
        Indices in ad.AnnData object for testing.

    Returns
    -------
    sort_idc : np.ndarray
        Ordered indices that will sort ad.AnnData object as all 
        training samples, then all testing.

    train_indices : np.ndarray
        The new training indices given the new index order, `sort_idc`.

    test_indices : np.ndarray
        The new testing indices given the new index order, `sort_idc`.
    """
    sort_idc = np.concatenate([train_indices, test_indices])

    train_indices = np.arange(0, train_indices.shape[0])
    test_indices = np.arange(train_indices.shape[0], 
                             train_indices.shape[0] + test_indices.shape[0])
    
    return sort_idc, train_indices, test_indices


def create_adata(X: scipy.sparse._csc.csc_matrix | np.ndarray | pd.DataFrame, 
                 feature_names: np.ndarray, cell_labels: np.ndarray, 
                 group_dict: dict, obs_names: None | np.ndarray=None, 
                 scale_data: bool=True, transform_data: bool=False, 
                 split_data: np.ndarray | None=None, D: int | None=None, 
                 remove_features: bool=True, train_ratio: float=0.8, 
                 distance_metric: str='euclidean', kernel_type: str='Gaussian', 
                 random_state: int=1, allow_multiclass: bool = False, 
                 class_threshold: str | int | None = None,
                 reduction: str | None = None, tfidf: bool = False):
    """
    Function to create an AnnData object to carry all relevant 
    information going forward.

    Parameters
    ----------
    X : scipy.sparse.csc_matrix | np.ndarray | pd.DataFrame
        A data matrix of cells by features (sparse array 
        recommended for large datasets).

    feature_names : np.ndarray
        Array of feature names corresponding with the features 
        in `X`.

    cell_labels : np.ndarray
        A numpy array of cell phenotypes corresponding with 
        the cells in `X`.

    group_dict : dict 
        Dictionary containing feature grouping information (i.e. 
        `{geneset1: np.array([gene_1, gene_2, ..., gene_n]), geneset2: 
        np.array([...]), ...}`.

    obs_names : None | np.ndarray
        The cell names corresponding to `X` to be assigned to output 
        object `.obs_names` attribute.

    scale_data : bool  
        If `True`, data matrix is log transformed and standard 
        scaled. Default is `True`.

    transform_data : bool
        If `True`, data will be log1p transformed (recommended for 
        counts data). Default is `False`.   
    
    split_data : None | np.ndarray
        If `None`, data will be split stratified by cell labels. 
        Else, is an array of precalculated train/test split 
        corresponding to samples. Can include labels for entire
        dataset to benchmark performance or for only training
        data to classify unknown cell types (i.e. `np.array(['train', 
        'test', ..., 'train'])`.

    D : int 
        Number of Random Fourier Features used to calculate Z. 
        Should be a positive integer. Higher values of D will 
        increase classification accuracy at the cost of computation 
        time. If set to `None`, will be calculated given number of 
        samples. 
    
    remove_features : bool
        If `True`, will remove features from `X` and `feature_names` 
        not in `group_dict` and remove features from groupings not in 
        `feature_names`.

    train_ratio : float
        Ratio of number of training samples to entire data set. Note:
        if a threshold is applied, the ratio training samples may 
        decrease depending on class balance and `class_threshold`
        parameter if `allow_multiclass = True`.

    distance_metric : str
        The pairwise distance metric used to estimate sigma. Must
        be one of the options used in `scipy.spatial.distance.cdist`.

    kernel_type : str
        The approximated kernel function used to calculate Zs.
        Must be one of `'Gaussian'`, `'Laplacian'`, or `'Cauchy'`.

    random_state : int
        Integer random_state used to set the seed for 
        reproducibilty.

    allow_multiclass : bool
        If `False`, will ensure that cell labels are binary.

    class_threshold : str | int
        Number of samples allowed in the training data for each cell
        class in the training data. If `'median'`, the median number 
        of cells per cell class will be the threshold for number of 
        samples per class.

    reduction: str | None
        Choose which dimension reduction technique to perform on 
        features within a group. 'svd' will run 
        `sklearn.decomposition.TruncatedSVD`, 'linear' will multiply 
        by an array of 1s down to 50 dimensions. 'pca' will replace 
        each group values with 50 PCs from principal component 
        analysis.
        
    tfidf: bool
        Whether to calculate TFIDF transformation on peaks within 
        groupings.
        
    Returns
    -------
    adata : ad.AnnData
        AnnData with the following attributes and keys:

        `adata.X` (array_like):
            Data matrix.
    
        `adata.var_names` (array_like): 
            Feature names corresponding to `adata.X`.

        `adata.obs['labels']` (array_like):
            cell classes/phenotypes from `cell_labels`.

        `adata.uns['train_indices']` (array_like):
            Indices for training data. 

        `adata.uns['test_indices']` (array_like)
            Indices for testing data.

        `adata.uns['group_dict']` (dict):
            Grouping information.

        `adata.uns['seed_obj']` (np.random._generator.Generator): 
            Seed object with seed equal to 100 * `random_state`.

        `adata.uns['D']` (int):
            Number of dimensions to scMKL with.

        `adata.uns['scale_data']` (bool):
            Whether or not data is scaled.

        `adata.uns['transform_data']` (bool):
            Whether or not data is log1p transformed.

        `adata.uns['distance_metric']` (str): 
            Distance metric as given.
    
        `adata.uns['kernel_type']` (str): 
            Kernel function as given.

        `adata.uns['svd']` (bool): 
            Whether to calculate SVD reduction.

        `adata.uns['tfidf']` (bool): 
            Whether to calculate TF-IDF per grouping.

    Examples
    --------
    >>> data_mat = scipy.sparse.load_npz('MCF7_RNA_matrix.npz')
    >>> gene_names = np.load('MCF7_gene_names.pkl', allow_pickle = True)
    >>> group_dict = np.load('hallmark_genesets.pkl', 
    >>>                      allow_pickle = True)
    >>> 
    >>> adata = scmkl.create_adata(X = data_mat, 
    ...                            feature_names = gene_names, 
    ...                            group_dict = group_dict)
    >>> adata
    AnnData object with n_obs × n_vars = 1000 × 4341
    obs: 'labels'
    uns: 'group_dict', 'seed_obj', 'scale_data', 'D', 'kernel_type', 
    'distance_metric', 'train_indices', 'test_indices'
    """

    assert X.shape[1] == len(feature_names), ("Different number of features "
                                              "in X than feature names")
    
    if not allow_multiclass:
        assert len(np.unique(cell_labels)) == 2, ("cell_labels must contain "
                                                  "2 classes")
    if D is not None:    
        assert isinstance(D, int) and D > 0, 'D must be a positive integer'

    kernel_options = ['gaussian', 'laplacian', 'cauchy']
    assert kernel_type.lower() in kernel_options, ("Given kernel type not "
                                                   "implemented. Gaussian, "
                                                   "Laplacian, and Cauchy "
                                                   "are the acceptable "
                                                   "types.")

    # Create adata object and add column names
    adata = ad.AnnData(X)
    adata.var_names = feature_names

    if isinstance(obs_names, (np.ndarray)):
        adata.obs_names = obs_names

    filtered_feature_names, group_dict = filter_features(feature_names, 
                                                          group_dict)
    
    # Ensuring that there are common features between feature_names and groups
    overlap_err = ("No common features between feature names and grouping "
                   "dict. Check grouping.")
    assert len(filtered_feature_names) > 0, overlap_err

    if remove_features:
        warnings.filterwarnings('ignore', category = ad.ImplicitModificationWarning)
        adata = adata[:, filtered_feature_names]
    
    gc.collect()

    # Add metadata to adata object
    adata.uns['group_dict'] = group_dict
    adata.uns['seed_obj'] = np.random.default_rng(100*random_state)
    adata.uns['scale_data'] = scale_data
    adata.uns['transform_data'] = transform_data
    adata.uns['D'] = D if D is not None else calculate_d(adata.shape[0])
    adata.uns['kernel_type'] = kernel_type
    adata.uns['distance_metric'] = distance_metric
    adata.uns['reduction'] = reduction if isinstance(reduction, str) else 'None'
    adata.uns['tfidf'] = tfidf

    if (split_data is None):
        assert X.shape[0] == len(cell_labels), ("Different number of cells "
                                                "than labels")
        adata.obs['labels'] = cell_labels

        if (allow_multiclass == False):
            split = binary_split(cell_labels, 
                                  seed_obj = adata.uns['seed_obj'],
                                  train_ratio = train_ratio)
            train_indices, test_indices = split

        elif (allow_multiclass == True):
            split = multi_class_split(cell_labels, 
                                       seed_obj = adata.uns['seed_obj'], 
                                       class_threshold = class_threshold,
                                       train_ratio = train_ratio)
            train_indices, test_indices = split

        adata.uns['labeled_test'] = True

    else:
        sd_err_message = "`split_data` argument must be of type np.ndarray"
        assert isinstance(split_data, np.ndarray), sd_err_message
        x_eq_labs = X.shape[0] == len(cell_labels)
        train_eq_labs = X.shape[0] == len(cell_labels)
        assert x_eq_labs or train_eq_labs, ("Must give labels for all cells "
                                            "or only for training cells")
        
        train_indices = np.where(split_data == 'train')[0]
        test_indices = np.where(split_data == 'test')[0]

        if len(cell_labels) == len(train_indices):

            padded_cell_labels = np.zeros((X.shape[0])).astype('object')
            padded_cell_labels[train_indices] = cell_labels
            padded_cell_labels[test_indices] = 'padded_test_label'

            adata.obs['labels'] = padded_cell_labels
            adata.uns['labeled_test'] = False

        elif len(cell_labels) == len(split_data):
            adata.obs['labels'] = cell_labels
            adata.uns['labeled_test'] = True

    # Ensuring all train samples are first in adata object followed by test
    sort_idx, train_indices, test_indices = sort_samples(train_indices, 
                                                         test_indices)

    adata = adata[sort_idx]

    if not isinstance(obs_names, (np.ndarray)):
        adata.obs = adata.obs.reset_index(drop=True)
        adata.obs.index = adata.obs.index.astype('O')

    adata.uns['train_indices'] = train_indices
    adata.uns['test_indices'] = test_indices

    if not scale_data:
        print("WARNING: Data will not be scaled. "
              "To change this behavior, set scale_data to True")
    if not transform_data:
        print("WARNING: Data will not be transformed."
              "To change this behavior, set transform_data to True")

    return adata


def format_adata(adata: ad.AnnData | str, cell_labels: np.ndarray | str, 
                 group_dict: dict | str, use_raw: bool=False, 
                 scale_data: bool=True, transform_data: bool=False, 
                 split_data: np.ndarray | None=None, D: int | None=None, 
                 remove_features: bool=True, train_ratio: float=0.8, 
                 distance_metric: str='euclidean', kernel_type: str='Gaussian', 
                 random_state: int=1, allow_multiclass: bool = False, 
                 class_threshold: str | int | None=None, 
                 reduction: str | None = None, tfidf: bool = False):
    """
    Function to format an `ad.AnnData` object to carry all relevant 
    information going forward. `adata.obs_names` will be retained.

    **NOTE: Information not needed for running `scmkl` will be 
    removed.**

    Parameters
    ----------
    adata : ad.AnnData
        Object with data for `scmkl` to be applied to. Only requirment 
        is that `.var_names` is correct and data matrix is in `adata.X` 
        or `adata.raw.X`. A h5ad file can be provided as a `str` and it 
        will be read in.

    cell_labels : np.ndarray | str
        If type `str`, the labels for `scmkl` to learn are captured 
        from `adata.obs['cell_labels']`. Else, a `np.ndarray` of cell 
        phenotypes corresponding with the cells in `adata.X`.

    group_dict : dict | str
        Dictionary containing feature grouping information (i.e. 
        `{geneset1: np.array([gene_1, gene_2, ..., gene_n]), geneset2: 
        np.array([...]), ...}`. A pickle file can be provided as a `str` 
        and it will be read in.

    obs_names : None | np.ndarray
        The cell names corresponding to `X` to be assigned to output 
        object `.obs_names` attribute.

    use_raw : bool
        If `False`, will use `adata.X` to create new `adata`. Else, 
        will use `adata.raw.X`.

    scale_data : bool  
        If `True`, data matrix is log transformed and standard 
        scaled. 

    transform_data : bool
        If `True`, data will be log1p transformed (recommended for 
        counts data). Default is `False`. 
        
    split_data : None | np.ndarray
        If `None`, data will be split stratified by cell labels. 
        Else, is an array of precalculated train/test split 
        corresponding to samples. Can include labels for entire
        dataset to benchmark performance or for only training
        data to classify unknown cell types (i.e. `np.array(['train', 
        'test', ..., 'train'])`.

    D : int 
        Number of Random Fourier Features used to calculate Z. 
        Should be a positive integer. Higher values of D will 
        increase classification accuracy at the cost of computation 
        time. If set to `None`, will be calculated given number of 
        samples. 
    
    remove_features : bool
        If `True`, will remove features from `X` and `feature_names` 
        not in `group_dict` and remove features from groupings not in 
        `feature_names`.

    train_ratio : float
        Ratio of number of training samples to entire data set. Note:
        if a threshold is applied, the ratio training samples may 
        decrease depending on class balance and `class_threshold`
        parameter if `allow_multiclass = True`.

    distance_metric : str
        The pairwise distance metric used to estimate sigma. Must
        be one of the options used in `scipy.spatial.distance.cdist`.

    kernel_type : str
        The approximated kernel function used to calculate Zs.
        Must be one of `'Gaussian'`, `'Laplacian'`, or `'Cauchy'`.

    random_state : int
        Integer random_state used to set the seed for 
        reproducibilty.

    allow_multiclass : bool
        If `False`, will ensure that cell labels are binary.

    class_threshold : str | int
        Number of samples allowed in the training data for each cell
        class in the training data. If `'median'`, the median number 
        of cells per cell class will be the threshold for number of 
        samples per class.

    reduction: str | None
        Choose which dimension reduction technique to perform on 
        features within a group. 'svd' will run 
        `sklearn.decomposition.TruncatedSVD`, 'linear' will multiply 
        by an array of 1s down to 50 dimensions.
        
    tfidf: bool
        Whether to calculate TFIDF transformation on peaks within 
        groupings.
        
    Returns
    -------
    adata : ad.AnnData
        AnnData with the following attributes and keys:

        `adata.X` (array_like):
            Data matrix.
    
        `adata.var_names` (array_like): 
            Feature names corresponding to `adata.X`.

        `adata.obs['labels']` (array_like):
            cell classes/phenotypes from `cell_labels`.

        `adata.uns['train_indices']` (array_like):
            Indices for training data. 

        `adata.uns['test_indices']` (array_like)
            Indices for testing data.

        `adata.uns['group_dict']` (dict):
            Grouping information.

        `adata.uns['seed_obj']` (np.random._generator.Generator): 
            Seed object with seed equal to 100 * `random_state`.

        `adata.uns['D']` (int):
            Number of dimensions to scMKL with.

        `adata.uns['scale_data']` (bool):
            Whether or not data is scaled.

        `adata.uns['transform_data']` (bool):
            Whether or not data is log1p transformed.

        `adata.uns['distance_metric']` (str): 
            Distance metric as given.
    
        `adata.uns['kernel_type']` (str): 
            Kernel function as given.

        `adata.uns['svd']` (bool): 
            Whether to calculate SVD reduction.

        `adata.uns['tfidf']` (bool): 
            Whether to calculate TF-IDF per grouping.

    Examples
    --------
    >>> adata = ad.read_h5ad('MCF7_rna.h5ad')
    >>> group_dict = np.load('hallmark_genesets.pkl', 
    >>>                      allow_pickle = True)
    >>> 
    >>> 
    >>> # The labels in adata.obs we want to learn are 'celltypes'
    >>> adata = scmkl.format_adata(adata, 'celltypes', 
    ...                            group_dict)
    >>> adata
    AnnData object with n_obs × n_vars = 1000 × 4341
    obs: 'labels'
    uns: 'group_dict', 'seed_obj', 'scale_data', 'D', 'kernel_type', 
    'distance_metric', 'train_indices', 'test_indices'
    """
    if str == type(adata):
        adata = ad.read_h5ad(adata)

    if str == type(group_dict):
        group_dict = np.load(group_dict, allow_pickle=True)
        
    if str == type(cell_labels):
        err_msg = f"{cell_labels} is not in `adata.obs`"
        assert cell_labels in adata.obs.keys(), err_msg
        cell_labels = adata.obs[cell_labels].to_numpy()
    
    if use_raw:
        assert adata.raw, "`adata.raw` is empty, set `use_raw` to `False`"
        X = adata.raw.X
    else:
        X = adata.X

    adata = create_adata(X, adata.var_names.to_numpy().copy(), cell_labels, 
                         group_dict, adata.obs_names.to_numpy().copy(), 
                         scale_data, transform_data, split_data, D, remove_features, 
                         train_ratio, distance_metric, kernel_type, 
                         random_state, allow_multiclass, class_threshold, 
                         reduction, tfidf)

    return adata



import scmkl
import os
import unittest
import numpy as np


def read_results(ftype='RNA', add_alpha_star=False):
    """
    Read scMKL results file.
    """
    if ftype != 'multiclass':
        file = f'../example/data/MCF7_RNA_results.pkl'
    else:
        file = '../example/data/celltype_results.pkl'

    file = np.load(file, allow_pickle=True)

    if add_alpha_star:
        file['Alpha_star'] = np.min(list(file['Norms'].keys()))

    return file


class TestPlotting(unittest.TestCase):
    """
    This unittest class is used to ensure that the plotting module is 
    running. Output plots for review will be placed in ./figures/.
    """
    def test_plot_conf_mat(self):
        """
        This function ensures that `scmkl.plot_conf_mat()` works 
        properly for both binary and multiclass results.
        """
        binary = read_results('RNA', True)
        multi = read_results('multiclass')

        scmkl.plot_conf_mat(binary, save='figures/plot_conf_mat_binary.png')
        scmkl.plot_conf_mat(multi, save='figures/plot_conf_mat_multi.png')


    def test_plot_metric(self):
        """
        This function ensures that `scmkl.plot_metric()` works 
        properly for both binary and multiclass results.
        """
        binary = read_results('RNA', True)
        multi = read_results('multiclass')

        binary_summary = scmkl.get_summary(binary)
        plot = scmkl.plot_metric(binary_summary, alpha_star=0.29)
        plot.save('figures/plot_metric_binary.png')

        multi_summary = scmkl.get_summary(multi)
        plot = scmkl.plot_metric(multi_summary)
        plot.save('figures/plot_metric_multi.png')

    
    def test_weights_barplot(self):
        """
        This function ensures that `scmkl.weights_barplot()` works 
        properly for both binary and multiclass results.
        """
        binary = read_results('RNA', True)
        multi = read_results('multiclass')

        binary_plot = scmkl.weights_barplot(binary, n_groups=50)
        binary_plot.save('figures/weights_barplot_binary.png')

        multi_plot = scmkl.weights_barplot(multi, n_groups=16)
        multi_plot.save('figures/weights_barplot_multi.png')


    def test_weights_heatmap(self):
        """
        This function ensures that `scmkl.weights_heatmap()` works 
        properly for both binary and multiclass results.
        """
        binary = read_results('RNA', True)
        multi = read_results('multiclass')

        binary_plot = scmkl.weights_heatmap(binary)
        binary_plot.save('figures/weights_heatmap_binary.png')

        multi_plot = scmkl.weights_heatmap(multi, scale_weights=True)
        multi_plot.save('figures/weights_heatmap_multi.png')


    def test_weights_dotplot(self):
        """
        This function ensures that `scmkl.weights_dotplot()` works 
        properly for both binary and multiclass results.
        """
        binary = read_results('RNA', True)
        multi = read_results('multiclass')

        binary_plot = scmkl.weights_dotplot(binary)
        binary_plot.save('figures/weights_dotplot_binary.png')

        multi_plot = scmkl.weights_heatmap(multi, scale_weights=True)
        multi_plot.save('figures/weights_dotplot_multi.png')


    def test_group_umap(self):
        """
        This function ensures that `scmkl.group_umap()` works 
        properly for both binary and multiclass results.
        """
        adata_fp = '../example/data/pbmc_{}.h5ad'
        group_fp = '../example/data/_{}_azimuth_pbmc_groupings.pkl'

        rna_adata = scmkl.format_adata(adata_fp.format('rna'), 
                                       'celltypes', 
                                       group_fp.format('RNA'), 
                                       allow_multiclass=True)
        scmkl.group_umap(rna_adata, g_name='B Cell Markers', 
                         title='B Cell Markers', save='_group_rna.png')
        
        atac_adata = scmkl.format_adata(adata_fp.format('atac'), 'celltypes',
                                        group_fp.format('ATAC'),
                                        allow_multiclass=True)

        scmkl.group_umap(atac_adata, g_name='CD16+ Monocyte Markers', 
                         is_binary=True, title='CD16+ Monocyte Markers', 
                         save='_group_atac.png')


if __name__ == '__main__':

    if not os.path.exists('figures/'):
        os.mkdir('figures/')

    unittest.main()


import numpy as np
import anndata as ad
import gc

from scmkl.tfidf_normalize import tfidf_normalize
from scmkl.data_processing import sparse_var
from scmkl.calculate_z import calculate_z


def combine_modalities(adatas: list[ad.AnnData], names: list[str], 
                        combination: str = 'concatenate'):
    """
    Combines data sets for multimodal classification. Combined group 
    names are `f'{assay}+{group_name}'`.

    Parameters
    ----------
    adatas : list[ad.AnnData]
        List of AnnData objects where each object is a different 
        modality. Annotations must match between objects (i.e. same 
        sample order).

    names : list[str]
        List of strings names for each modality repective to each 
        object in adatas.
            
    combination: str
        How to combine the matrices, either `'sum'` or `'concatenate'`.
    
    Returns
    -------
    combined_adata : ad.Anndata
        Adata object with the combined Z matrices and annotations. 
    """
    assert len({adata.shape[0] for adata in adatas}) == 1, ("All adatas must "
                                                            "have the same "
                                                            "number of rows")
    assert len(np.unique(names)) == len(names), "Assay names must be distinct"
    assert combination.lower() in ['sum', 'concatenate']

    z_train = all(['Z_train' in adata.uns.keys() for adata in adatas])
    z_test = all(['Z_test' in adata.uns.keys() for adata in adatas])

    assert all([z_train, z_test]), "Z not calculated for one or more adatas"

    # Combining modalities
    combined_adata = ad.concat(adatas, uns_merge = 'same', 
                               axis = 1, label = 'labels')
    
    assert 'train_indices' in combined_adata.uns.keys(), ("Different train "
                                                          "test splits "
                                                          "between AnnData "
                                                          "objects")

    # Conserving labels from adatas
    combined_adata.obs = adatas[0].obs.copy()

    # Creating a single dictionary with all of the groups across modalities 
    group_dict = {}
    for name, adata in zip(names, adatas):
        for group_name, features in adata.uns['group_dict'].items():
            group_dict[f'{name}-{group_name}'] = features

    if combination == 'concatenate':
        combined_adata.uns['Z_train'] = np.hstack([adata.uns['Z_train'] 
                                                   for adata in adatas])
        combined_adata.uns['Z_test'] = np.hstack([adata.uns['Z_test'] 
                                                  for adata in adatas])


    elif combination == 'sum':

        #Check that the dimensions of all Z's are the same
        dims = [adata.uns['Z_train'].shape for adata in adatas]
        dims = all([dim == dims[0] for dim in dims])
        assert dims, "Cannot sum Z matrices with different dimensions"
        
        combined_adata.uns['Z_train'] = np.sum([adata.uns['Z_train'] 
                                                for adata in adatas], 
                                                axis = 0)
        combined_adata.uns['Z_test'] = np.sum([adata.uns['Z_test'] 
                                               for adata in adatas], 
                                               axis = 0)


    combined_adata.uns['group_dict'] = group_dict

    if 'seed_obj' in adatas[0].uns_keys():
        combined_adata.uns['seed_obj'] = adatas[0].uns['seed_obj']
    else:
        print("No random seed present in adata"
              "Recommended for reproducibility.")

    del adatas
    gc.collect()

    return combined_adata


def multimodal_processing(adatas : list[ad.AnnData], names : list[str], 
                          tfidf: list[bool], combination: str='concatenate', 
                          batches: int=10, batch_size: int=100, 
                          verbose: bool=True) -> ad.AnnData:
    """
    Combines and processes a list of `ad.AnnData` objects.

    Parameters
    ----------
    adatas : list[ad.AnnData]
        List of `ad.AnnData` objects where each object is a different 
        modality. Annotations must match between objects (i.e. same 
        sample order).

    names : list[str]
        List of strings names for each modality repective to each 
        object in adatas.
            
    combination: str
        How to combine the matrices, either `'sum'` or `'concatenate'`.
    
    tfidf : list[bool]
        If element `i` is `True`, `adata[i]` will be TF-IDF normalized.

    batches : int
        The number of batches to use for the distance calculation.
        This will average the result of `batches` distance calculations
        of `batch_size` randomly sampled cells. More batches will converge
        to population distance values at the cost of scalability.

    batch_size : int
        The number of cells to include per batch for distance
        calculations. Higher batch size will converge to population
        distance values at the cost of scalability.
        If `batches*batch_size > num_training_cells`, `batch_size` 
        will be reduced to `int(num_training_cells / batches)`.

    Returns
    -------
    adata : ad.AnnData
        Concatenated from objects from `adatas` with Z matrices 
        calculated.

    Examples
    --------
    >>> rna_adata = scmkl.create_adata(X = mcf7_rna_mat, 
    ...                                feature_names=gene_names, 
    ...                                scale_data=True, 
    ...                                transform_data=True,
    ...                                cell_labels=cell_labels, 
    ...                                 group_dict=rna_grouping)
    >>>
    >>> atac_adata = scmkl.create_adata(X=mcf7_atac_mat, 
    ...                                 feature_names=peak_names, 
    ...                                 scale_data=False, 
    ...                                 cell_labels=cell_labels, 
    ...                                 group_dict=atac_grouping)
    >>>
    >>> adatas = [rna_adata, atac_adata]
    >>> mod_names = ['rna', 'atac']
    >>> adata = scmkl.multimodal_processing(adatas = adatas, 
    ...                                     names = mod_names,
    ...                                     tfidf = [False, True])
    >>>
    >>> adata
    AnnData object with n_obs × n_vars = 1000 × 12676
    obs: 'labels'
    var: 'labels'
    uns: 'D', 'kernel_type', 'distance_metric', 'train_indices',  
    'test_indices', 'Z_train', 'Z_test', 'group_dict', 'seed_obj'
    """
    import warnings 
    warnings.filterwarnings('ignore')

    diff_num_warn = "Different number of cells present in each object."
    assert all([adata.shape[0] for adata in adatas]), diff_num_warn
    
    # True if all train indices match
    same_train = np.all([np.array_equal(adatas[0].uns['train_indices'], 
                                        adatas[i].uns['train_indices']) 
                         for i in range(1, len(adatas))])

    # True if all test indices match
    same_test = np.all([np.array_equal(adatas[0].uns['test_indices'], 
                                       adatas[i].uns['test_indices']) 
                        for i in range(1, len(adatas))])

    assert same_train, "Different train indices"
    assert same_test, "Different test indices"

    # Creates a boolean array for each modality of cells with non-empty rows
    non_empty_rows = [np.array(sparse_var(adata.X, axis = 1) != 0).ravel() 
                      for adata in adatas]
    non_empty_rows = np.transpose(non_empty_rows)

    # Returns a 1D array where sample feature sums non-0 across all modalities
    non_empty_rows = np.array([np.all(non_empty_rows[i])
                              for i in range(non_empty_rows.shape[0])])

    # Initializing final train test split array
    train_test = np.repeat('train', adatas[0].shape[0])
    train_test[adatas[0].uns['test_indices']] = 'test'

    # Capturing train test split with empty rows filtered out
    train_test = train_test[non_empty_rows]
    train_indices = np.where(train_test == 'train')[0]
    test_indices = np.where(train_test == 'test')[0]

    # Adding train test split arrays to AnnData objects 
    # and filtering out empty samples
    for i, adata in enumerate(adatas):
        adatas[i].uns['train_indices'] = train_indices
        adatas[i].uns['test_indices'] = test_indices
        adatas[i] = adata[non_empty_rows, :]

        # tfidf normalizing if corresponding element in tfidf is True
        if tfidf[i]:
            adatas[i] = tfidf_normalize(adata)
        
        if verbose:
            print(f"Estimating sigma and calculating Z for {names[i]}", 
                  flush = True)
        adatas[i] = calculate_z(adata, n_features = 5000, batches=batches, 
                                batch_size=batch_size)

    if 'labels' in adatas[0].obs:
        all_labels = [adata.obs['labels'] for adata in adatas]
        # Ensuring cell labels for each AnnData object are the same
        uneq_labs_warn = ("Cell labels between AnnData object in position 0 "
                          "and position {} in adatas do not match")
        for i in range(1, len(all_labels)):
            same_labels = np.all(all_labels[0] == all_labels[i])
            assert same_labels, uneq_labs_warn.format(i)

    adata = combine_modalities(adatas=adatas,
                                names=names,
                                combination=combination)

    del adatas
    gc.collect()

    return adata    



import scmkl
import unittest
import numpy as np
from test_calculate_z import create_test_adata


class TestOneVRest(unittest.TestCase):
    """
    This unittest class is used to ensure that scmkl.one_v_rest() 
    works properly for both unimodal and multimodal runs.
    """
    def test_unimodal_one_v_rest(self):
        """
        This function ensure that scmkl.one_v_rest() works correctly 
        for unimodal test cases by checking that the output is what we 
        expect without failing.
        """
        # Creating adata and setting alphas to choose from
        adata = create_test_adata('RNA')
        alpha_list = np.array([0.1, 0.2])

        # Creating dummy labels and train/test split to test function
        ct = ['B', 'T', 'Mono']
        train_idc = np.arange(0, 800)
        test_idc = np.arange(800, 1000)

        adata.obs['labels'] = [ct[i%3] for i in range(adata.shape[0])]
        adata.uns['train_indices'] = train_idc
        adata.uns['test_indices'] = test_idc

        adata = scmkl.calculate_z(adata, batch_size=80)
        results = scmkl.one_v_rest([adata], ['rna'], 
                                   alpha_params=alpha_list, 
                                   tfidf=[False],
                                   batch_size=80)

        expected_keys = np.array(ct + ['Macro_F1-Score', 
                                       'Predicted_class', 
                                       'Truth_labels'])
        keys_in = np.array([key in results.keys() for key in expected_keys])
        
        self.assertTrue(np.all(keys_in), 
                        f"Expect keys missing: {expected_keys[~keys_in]}")

if __name__ == '__main__':
    unittest.main()


import numpy as np
import scipy
from sklearn.decomposition import TruncatedSVD, PCA
import anndata as ad


def sparse_var(X: scipy.sparse._csc.csc_matrix | np.ndarray, axis: int | None=None):
    """
    Function to calculate variance on a scipy sparse matrix.
    
    Parameters
    ----------
    X : scipy.sparse._csc.csc_matrix | np.ndarray
        A scipy sparse or numpy array
        
    axis : int | None
        Determines which axis variance is calculated on. Same usage 
        as Numpy.
    
    Returns
    -------
    var : np.ndarray | float
        Variance values calculated over the given axis.
    """
    # E[X^2] - E[X]^2
    if scipy.sparse.issparse(X):
        exp_mean = np.asarray(X.power(2).mean(axis = axis)).flatten()
        sq_mean = np.asarray(np.square(X.mean(axis = axis))).flatten()
        var = np.array(exp_mean - sq_mean)
    else:
        var = np.asarray(np.var(X, axis = axis)).flatten()

    return var.ravel()


def process_data(X_train: np.ndarray | scipy.sparse._csc.csc_matrix,
                 X_test: np.ndarray | scipy.sparse._csc.csc_matrix | None=None,
                 scale_data: bool=True, transform_data: bool=False,
                 return_dense: bool=True):
    """
    Function to preprocess data matrix according to type of data 
    (e.g. counts/rna, or binary/atac). Will process test data 
    according to parameters calculated from test data.
    
    Parameters
    ----------
    X_train : np.ndarray | scipy.sparse._csc.csc_matrix
        A scipy sparse or numpy array of cells x features in the 
        training data.

    X_test : np.ndarray | scipy.sparse._csc.csc_matrix
        A scipy sparse or numpy array of cells x features in the 
        testing data.

    scale_data : bool
        If `True`, data will be logarithmized then z-score 
        transformed.

    transform_data : bool
        If `True`, data will be log1p transformed (recommended for 
        counts data). Default is `False`.

    return_dense: bool
        If `True`, a np.ndarray will be returned as opposed to a 
        scipy.sparse object.
    
    Returns
    -------
    X_train, X_test : np.ndarray, np.ndarray
        Numpy arrays with the process train/test data 
        respectively. If X_test is `None`, only X_train is returned.
    """
    if X_test is None:
        # Creates dummy matrix to for the sake of calculation without 
        # increasing computational time
        X_test = X_train[:1,:] 
        orig_test = None
    else:
        orig_test = 'given'

    # Remove features that have no variance in the training data 
    # (will be uniformative)
    var = sparse_var(X_train, axis = 0)
    variable_features = np.where(var > 1e-5)[0]

    X_train = X_train[:,variable_features]
    X_test = X_test[:, variable_features]

    # Data processing according to data type
    if transform_data:

        if scipy.sparse.issparse(X_train):
            X_train = X_train.log1p()
            X_test = X_test.log1p()
        else:
            X_train = np.log1p(X_train)
            X_test = np.log1p(X_test)
    
    if scale_data:
        #Center and scale count data
        train_means = np.mean(X_train, 0)
        train_sds = np.sqrt(var[variable_features])

        # Perform transformation on test data according to parameters 
        # of the training data
        X_train = (X_train - train_means) / train_sds
        X_test = (X_test - train_means) / train_sds


    if return_dense and scipy.sparse.issparse(X_train):
        X_train = X_train.toarray()
        X_test = X_test.toarray()


    if orig_test is None:
        return X_train
    else:
        return X_train, X_test
    

def svd_transformation(X_train: scipy.sparse._csc.csc_matrix | np.ndarray,
                       X_test: scipy.sparse._csc.csc_matrix | 
                       np.ndarray | None=None):
    """
    Returns matrices with SVD reduction. If `X_test is None`, only 
    X_train is returned.

    Parameters
    ----------
    X_train : np.ndarray
        A 2D array of cells x features filtered to desired features 
        for training data.

    X_test : np.ndarray | None
        A 2D array of cells x features filtered to desired features 
        for testing data.
    
    Returns
    -------
    X_train, X_test : np.ndarray, np.ndarray
        Transformed matrices. Only X_train is returned if 
        `X_test is None`.
    """
    n_components = np.min([50, X_train.shape[1]])
    SVD_func = TruncatedSVD(n_components = n_components, random_state = 1)
    
    # Remove first component as it corresponds with sequencing depth
    # We convert to a csr_array because the SVD function is faster on this
    # matrix type
    X_train = SVD_func.fit_transform(scipy.sparse.csr_array(X_train))[:, 1:]

    if X_test is not None:
        X_test = SVD_func.transform(scipy.sparse.csr_array(X_test))[:, 1:]
    
    return X_train, X_test


def sample_cells(train_indices: np.ndarray,
                 sample_size: int,
                 seed_obj: np.random._generator.Generator):
    """
    Samples cells indices from training indices for calculations.

    Parameters
    ----------
    train_indices : np.ndarray
        An array of indices to sample from.

    sample_size : int
        Number of samples to take from `train_indices`. Must be 
        smaller than length of `train_indices`.

    Returns
    -------
    indices : np.ndarray
        The sampled indices from `train_indices`.
    """
    n_samples = np.min((train_indices.shape[0], sample_size))
    indices = seed_obj.choice(train_indices, n_samples, replace = False)

    return indices


def pca_transformation(X_train: scipy.sparse._csc.csc_matrix | np.ndarray,
                       X_test: scipy.sparse._csc.csc_matrix | np.ndarray | None=None):
    """
    Returns matrices with PCA reduction. If `X_test is None`, only 
    X_train is returned.

    Parameters
    ----------
    X_train : scipy.sparse._csc.csc_matrix | np.ndarray
        A 2D array of cells x features filtered to desired features 
        for training data.

    X_test : scipy.sparse._csc.csc_matrix | np.ndarray | None
        A 2D array of cells x features filtered to desired features 
        for testing data.
    
    Returns
    -------
    X_train, X_test : np.ndarray, np.ndarray
        Transformed matrices. Only X_train is returned if 
        `X_test is None`.
    """
    n_components = np.min([50, X_train.shape[1]])
    PCA_func = PCA(n_components = n_components, random_state = 1)

    X_train = PCA_func.fit_transform(np.asarray(X_train))

    if X_test is not None:
        X_test = PCA_func.transform(np.asarray(X_test))
    
    return X_train, X_test


def _no_transformation(X_train: scipy.sparse._csc.csc_matrix | np.ndarray,
                      X_test: scipy.sparse._csc.csc_matrix | np.ndarray | None=None):
    """
    Dummy function used to return mat inputs.
    """
    return X_train, X_test


def get_reduction(reduction: str):
    """
    Function used to identify reduction type and return function to 
    apply to data matrices.

    Parameters
    ----------
    reduction : str
        The reduction for data transformation. Options are `['pca', 
        'svd', 'None']`.

    Returns
    -------
    red_func : function
        The function to reduce the data.
    """
    match reduction:
        case 'pca':
            red_func = pca_transformation
        case 'svd':
            red_func = svd_transformation
        case 'None':
            red_func = _no_transformation

    return red_func


def get_group_mat(adata: ad.AnnData, n_features: int,
                  group_features: np.ndarray,
                  n_group_features: int, 
                  process_test: bool=False) -> np.ndarray:
    """
    Filters to only features in group. Will sample features if 
    `n_features < n_group_features`.

    Parameters
    ----------
    adata : anndata.AnnData
        anndata object with `'seed_obj'`, `'train_indices'`, and 
        `'test_indices'` in `.uns`.

    n_features : int
        Maximum number of features to keep in matrix. Only 
        impacts mat if `n_features < n_group_features`.
    
    group_features : list | tuple | np.ndarray
        Feature names in group to filter matrices to.

    n_group_features : int
        Number of features in group.

    n_samples : int
        Number of samples to filter X_train to.

    Returns
    -------
    X_train, X_test : np.ndarray, np.ndarray
        Filtered matrices. If `n_samples` is provided, only `X_train` 
        is returned. If `adata.uns['reduction']` is `'pca'` or 
        `'svd'` the matrices are transformed before being returned.
    """
    # Getting reduction function
    reduction_func = get_reduction(adata.uns['reduction'])

    # Sample up to n_features features- important for scalability if 
    # using large groupings
    # Will use all features if the grouping contains fewer than n_features
    number_features = np.min([n_features, n_group_features])
    group_array = np.array(list(group_features))
    group_features = adata.uns['seed_obj'].choice(group_array, 
                                                  number_features, 
                                                  replace = False) 

    # Create data arrays containing only features within this group
    if process_test:
        X_train = adata[adata.uns['train_indices'],:][:, group_features].X
        X_test = adata[adata.uns['test_indices'],:][:, group_features].X
        X_train, X_test = reduction_func(X_train, X_test)
        return X_train, X_test

    else:
        X_train = adata[:, group_features].X
        return X_train


import scmkl
import unittest
import numpy as np


class TestGetRegionGrouping(unittest.TestCase):
    """
    This unittest class is used to ensure that 
    scmkl.get_region_groupings() works properly.
    """

    def test_get_region_groupings(self):
        """
        This function ensures that scmkl.get_region_grouping() works 
        as expected.
        """
        # Reading in gtf file
        gtf_fp = '../example/data/_hg38_subset_protein_coding.annotation.gtf'
        gtf = scmkl.read_gtf(gtf_fp, filter_to_coding=True)
        
        # Reading in gene sets and feature names
        gene_sets = np.load('../example/data/_RNA_hallmark_groupings.pkl',
                            allow_pickle=True)
        regions = np.load('../example/data/_MCF7_ATAC_feature_names.npy', 
                          allow_pickle=True)
        
        # Creating epigenetic grouping, and empty grouping
        region_grouping = scmkl.get_region_groupings(gtf, gene_sets, regions)
        empty_grouping = scmkl.get_region_groupings(gtf, gene_sets, regions, 
                                                    len_up=10, len_down=10)

        # Reading in larger, complete peak set to check output
        truth_grouping = np.load('../example/data/_ATAC_hallmark_groupings.pkl', 
                                 allow_pickle=True)
        
        # Check output grouping
        self.assertEqual(len(region_grouping), 50, 
                         "Incorrect number of groupings, should be 50.")
        
        for group, regions in region_grouping.items():
            for region in regions:
                self.assertIn(region, truth_grouping[group], 
                              f"Feature {region} should not be in {group}")
                
        # Ensuring no peaks matched genes with such small promoter regions
        empty_sizes = [len(empty_grouping[group]) <= 2
                       for group in empty_grouping.keys()]
        self.assertTrue(np.all(empty_sizes), "Group sizes should be minimal.")                      


if __name__ == '__main__':
    unittest.main()


import os
import re
import numpy as np
import pandas as pd


def _parse_result_type(results: dict):
    """
    Returns whether or not there are multiple results per class.

    Parameters
    ----------
    results : dict
        Either the output of `scmkl.run()` or `scmkl.one_v_rest()` or 
        a dictionary of those results.

    Returns
    -------
    is_mult, is_many : bool, bool
        If `is_mult` is `True`, then results are multiclass. If 
        `is_many` is `True`, results contain multiple outputs.

    """
    # Single result cases
    if 'Classes' in results.keys():
        is_mult = True
        is_many = False
        return is_mult, is_many
    elif 'Norms' in results.keys():
        is_mult = False
        is_many = False
        return is_mult, is_many

    # Multiresult cases
    keys = list(results.keys())
    if 'Classes' in results[keys[0]].keys():
        is_mult = True
        is_many = True
        return is_mult, is_many
    elif 'Norms' in results[keys[0]].keys():
        is_mult = False
        is_many = True
        return is_mult, is_many
    else:
        print("Unknown result structure", flush=True)


def sort_groups(df: pd.DataFrame, group_col: str='Group', 
                norm_col: str='Kernel Weight'):
    """
    Takes a dataframe with `group_col` and returns sorted group list 
    with groups in decending order by their weights. Assumes there is 
    one instance of each group.

    Parameters
    ----------
    df : pd.DataFrame
        A dataframe with `group_col` and `norm_col` to be sorted by.

    group_col : str
        The column containing the group names.

    norm_col : str
        The column containing the kernel weights.

    Returns
    -------
    group_order : list
        A list of groups in descending order according to their kernel 
        weights.

    Examples
    --------
    >>> result = scmkl.run(adata, alpha_list)
    >>> weights = scmkl.get_weights(result)
    >>> group_order = scmkl.sort_groups(weights, 'Group', 
    ...                                 'Kernel Weight')
    >>>
    >>> group_order
    ['HALLMARK_ESTROGEN_RESPONSE_EARLY', 'HALLM...', ...]
    """
    df = df.copy()
    df = df.sort_values(norm_col, ascending=False)
    group_order = list(df[group_col])

    return group_order


def format_group_names(group_names: list | pd.Series | np.ndarray, 
                       rm_words: list=list()):
    """
    Takes an ArrayLike object of group names and formats them.

    Parameters
    ----------
    group_names : array_like
        An array of group names to format.

    rm_words : list
        Words to remove from all group names.

    Returns
    -------
    new_group_names : list
        Formatted version of the input group names.

    Examples
    --------
    >>> groups = ['HALLMARK_E2F_TARGETS', 'HALLMARK_HYPOXIA']
    >>> new_groups = scmkl.format_group_names(groups)
    >>> new_groups
    ['Hallmark E2F Targets', 'Hallmark Hypoxia']
    """
    new_group_names = list()
    rm_words = [word.lower() for word in rm_words]

    for name in group_names:
        new_name = list()
        for word in re.split(r'_|\s', name):
            if word.isalpha() and (len(word) > 3):
                word = word.capitalize()
            if word.lower() not in rm_words:
                new_name.append(word)
        new_name = ' '.join(new_name)
        new_group_names.append(new_name)

    return new_group_names
        

def parse_metrics(results: dict, key: str | None=None, 
                   include_as: bool=False) -> pd.DataFrame:
    """
    This function returns a pd.DataFrame for a single scMKL result 
    with performance results.

    Parameters
    ----------
    results : dict
        A result dictionary from `scmkl.run()`.
    
    key : str
        If specified, will add a key column to the output dataframe 
        where each element is `key`.

    include_as : bool
        If `True`, will add a column indicating which models' used 
        the optimal alphas.

    Returns
    -------
    df : pd.DataFrame
        A dataframe with columns `['Alpha', 'Metric', 'Value']`. 
        `'Key'` col only added if `key` is not `None`.
    """
    df = {
        'Alpha' : list(),
        'Metric' : list(),
        'Value' : list()
    }

    # Check if is a multiclass result
    is_mult, _ = _parse_result_type(results)

    if is_mult:
        df['Class'] = list()

    # Ensuring results is a scMKL result and checking multiclass
    if 'Metrics' in results.keys():
        for alpha in results['Metrics'].keys():
            for metric, value in results['Metrics'][alpha].items():
                df['Alpha'].append(alpha)
                df['Metric'].append(metric)
                df['Value'].append(value)

    elif 'Classes' in results.keys():
        for ct in results['Classes']:
            for alpha in results[ct]['Metrics'].keys():
                for metric, value in results[ct]['Metrics'][alpha].items():
                    df['Alpha'].append(alpha)
                    df['Metric'].append(metric)
                    df['Value'].append(value)
                    df['Class'].append(ct)

    else:
        print(f"{key} is not a scMKL result and will be ignored.")
            
    df = pd.DataFrame(df)
    
    if include_as:
        assert 'Alpha_star' in results.keys(), "'Alpha_star' not in results"
        df['Alpha Star'] = df['Alpha'] == results['Alpha_star']

    if key is not None:
        df['Key'] = [key] * df.shape[0]

    return df        


def parse_weights(results: dict, include_as: bool=False, 
                   key: None | str=None) -> pd.DataFrame:
    """
    This function returns a pd.DataFrame for a single scMKL result 
    with group weights.

    Parameters
    ----------
    results : dict
        A result dictionary from `scmkl.run()`.
    
    key : str
        If specified, will add a key column to the output dataframe 
        where each element is `key`.

    include_as : bool
        If `True`, will add a column indicating which models' used 
        the optimal alphas.

    Returns
    -------
    df : pd.DataFrame
        A dataframe with columns `['Alpha', 'Group', 
        'Kernel Weight']`. `'Key'` col only added if `key` is not 
        `None`.
    """
    df = {
        'Alpha' : list(),
        'Group' : list(),
        'Kernel Weight' : list()
    }

    # Check if is a multiclass result
    is_mult, _ = _parse_result_type(results)

    if is_mult:
        df['Class'] = list()

    # Ensuring results is a scMKL result and checking multiclass
    if 'Norms' in results.keys():
        for alpha in results['Norms'].keys():
            df['Alpha'].extend([alpha]*len(results['Norms'][alpha]))
            df['Group'].extend(results['Group_names'])
            df['Kernel Weight'].extend(results['Norms'][alpha])

    elif 'Classes' in results.keys():
        for ct in results['Classes']:
            for alpha in results[ct]['Norms'].keys():
                df['Alpha'].extend([alpha] * len(results[ct]['Norms'][alpha]))
                df['Group'].extend(results[ct]['Group_names'])
                df['Kernel Weight'].extend(results[ct]['Norms'][alpha])
                df['Class'].extend([ct]*len(results[ct]['Norms'][alpha]))

    df = pd.DataFrame(df)
    
    if include_as:
        df['Alpha Star'] = df['Alpha'] == results['Alpha_star'] 

    if key is not None:
        df['Key'] = [key] * df.shape[0]

    return df


def extract_results(results: dict, metric: str):
    """
    
    """
    summary = {'Alpha' : list(),
               metric : list(),
               'Number of Selected Groups' : list(),
               'Top Group' : list()}
    
    alpha_list = list(results['Metrics'].keys())

    # Creating summary DataFrame for each model
    for alpha in alpha_list:
        cur_alpha_rows = results['Norms'][alpha]
        top_weight_rows = np.max(results['Norms'][alpha])
        top_group_index = np.where(cur_alpha_rows == top_weight_rows)
        num_selected = len(results['Selected_groups'][alpha])
        top_group_name = np.array(results['Group_names'])[top_group_index]
        
        if 0 == num_selected:
            top_group_name = ["No groups selected"]

        summary['Alpha'].append(alpha)
        summary[metric].append(results['Metrics'][alpha][metric])
        summary['Number of Selected Groups'].append(num_selected)
        summary['Top Group'].append(*top_group_name)
    
    return pd.DataFrame(summary)


def get_summary(results: dict, metric: str='AUROC'):
    """
    Takes the results from `scmkl.run()` and generates a dataframe 
    for each model containing columns for alpha, area under the ROC, 
    number of groups with nonzero weights, and highest weighted 
    group.

    Parameters
    ----------
    results : dict
        A dictionary of results from scMKL generated from 
        `scmkl.run()`.

    metric : str
        Which metric to include in the summary. Default is AUROC. 
        Options include `'AUROC'`, `'Recall'`, `'Precision'`, 
        `'Accuracy'`, and `'F1-Score'`.

    Returns
    -------
    summary_df : pd.DataFrame
        A table with columns: `['Alpha', 'AUROC', 
        'Number of Selected Groups', 'Top Group']`.
    
    Examples
    --------
    >>> results = scmkl.run(adata, alpha_list)
    >>> summary_df = scmkl.get_summary(results)
    ...
    >>> summary_df.head()
        Alpha   AUROC  Number of Selected Groups 
    0   2.20  0.8600                          3   
    1   1.96  0.9123                          4   
    2   1.72  0.9357                          5   
    3   1.48  0.9524                          7   
    4   1.24  0.9666                          9   
        Top Group
    0   RNA-HALLMARK_E2F_TARGETS
    1   RNA-HALLMARK_ESTROGEN_RESPONSE_EARLY
    2   RNA-HALLMARK_ESTROGEN_RESPONSE_EARLY
    3   RNA-HALLMARK_ESTROGEN_RESPONSE_EARLY
    4   RNA-HALLMARK_ESTROGEN_RESPONSE_EARLY
    """
    is_multi, is_many = _parse_result_type(results)
    assert not is_many, "This function only supports single results"
    
    if is_multi:
        summaries = list()
        for ct in results['Classes']:
            data = extract_results(results[ct], metric)
            data['Class'] = [ct]*len(data)
            summaries.append(data.copy())
        summary = pd.concat(summaries)

    else:
        summary = extract_results(results, metric)

    return summary


def read_files(dir: str, pattern: str | None=None) -> dict:
    """
    This function takes a directory of scMKL results as pickle files 
    and returns a dictionary with the file names as keys and the data 
    from the respective files as the values.

    Parameters
    ----------
    dir : str
        A string specifying the file path for the output scMKL runs.

    pattern : str
        A regex string for filtering down to desired files. If 
        `None`, all files in the directory with the pickle file 
        extension will be added to the dictionary.

    Returns
    -------
    results : dict
        A dictionary with the file names as keys and data as values.

    Examples
    --------
    >>> filepath = 'scMKL_results/rna+atac/'
    ...
    >>> all_results = scmkl.read_files(filepath)
    >>> all_results.keys()
    dict_keys(['Rep_1.pkl', Rep_2.pkl, Rep_3.pkl, ...])
    """
    # Reading all pickle files in patter is None
    if pattern is None:
        data = {file : np.load(f'{dir}/{file}', allow_pickle = True)
                 for file in os.listdir(dir) if '.pkl' in file}
    
    # Reading only files matching pattern if not None
    else:
        pattern = repr(pattern)
        data = {file : np.load(f'{dir}/{file}', allow_pickle = True)
                 for file in os.listdir(dir) 
                 if re.fullmatch(pattern, file) is not None}
        
    return data


def get_metrics(results: dict, include_as: bool=False) -> pd.DataFrame:
    """
    Takes either a single scMKL result or a dictionary where each 
    entry cooresponds to one result. Returns a dataframe with cols 
    ['Alpha', 'Metric', 'Value']. If `include_as == True`, another 
    col of booleans will be added to indicate whether or not the run 
    respective to that alpha was chosen as optimal via CV. If 
    `include_key == True`, another column will be added with the name 
    of the key to the respective file (only applicable with multiple 
    results).

    Parameters
    ----------
    results : dict | None
        A dictionary with the results of a single run from 
        `scmkl.run()`. Must be `None` if `rfiles is not None`.

    rfiles : dict | None
        A dictionary of results dictionaries containing multiple 
        results from `scmkl.run()`. 

    include_as : bool
        When `True`, will add a bool col to output pd.DataFrame 
        where rows with alphas cooresponding to alpha_star will be 
        `True`.

    Returns
    -------
    df : pd.DataFrame
        A pd.DataFrame containing all of the metrics present from 
        the runs input.

    Examples
    --------
    >>> # For a single file
    >>> results = scmkl.run(adata)
    >>> metrics = scmkl.get_metrics(results = results)

    >>> # For multiple runs saved in a dict
    >>> output_dir = 'scMKL_outputs/'
    >>> rfiles = scmkl.read_files(output_dir)
    >>> metrics = scmkl.get_metrics(rfiles=rfiles)
    """
    # Checking which data is being worked with 
    is_mult, is_many = _parse_result_type(results)

    # Initiating col list with minimal columns
    cols = ['Alpha', 'Metric', 'Value']

    if include_as:
        cols.append('Alpha Star')
    if is_mult:
        cols.append('Class')

    if is_many:
        cols.append('Key')
        df = pd.DataFrame(columns = cols)
        for key, result in results.items():
                cur_df = parse_metrics(results = result, key = key, 
                                        include_as = include_as)
                df = pd.concat([df, cur_df.copy()])
            
    else:
        df = parse_metrics(results = results, include_as = include_as)

    return df


def get_weights(results : dict, include_as : bool = False) -> pd.DataFrame:
    """
    Takes either a single scMKL result or dictionary of results and 
    returns a pd.DataFrame with cols ['Alpha', 'Group', 
    'Kernel Weight']. If `include_as == True`, a fourth col will be 
    added to indicate whether or not the run respective to that alpha 
    was chosen as optimal via cross validation.

    Parameters
    ----------
    results : dict | None
        A dictionary with the results of a single run from 
        `scmkl.run()`. Must be `None` if `rfiles is not None`.

    rfiles : dict | None
        A dictionary of results dictionaries containing multiple 
        results from `scmkl.run()`. 

    include_as : bool
        When `True`, will add a bool col to output pd.DataFrame 
        where rows with alphas cooresponding to alpha_star will be 
        `True`.

    Returns
    -------
    df : pd.DataFrame
        A pd.DataFrame containing all of the groups from each alpha 
        and their cooresponding kernel weights.

    Examples
    --------
    >>> # For a single file
    >>> results = scmkl.run(adata)
    >>> weights = scmkl.get_weights(results = results)
    
    >>> # For multiple runs saved in a dict
    >>> output_dir = 'scMKL_outputs/'
    >>> rfiles = scmkl.read_files(output_dir)
    >>> weights = scmkl.get_weights(rfiles=rfiles)
    """
    # Checking which data is being worked with 
    is_mult, is_many = _parse_result_type(results)

    # Initiating col list with minimal columns
    cols = ['Alpha', 'Group', 'Kernel Weight']

    if include_as:
        cols.append('Alpha Star')
    if is_mult:
        cols.append('Class')

    if is_many:
        cols.append('Key')
        df = pd.DataFrame(columns = cols)
        for key, result in results.items():
            cur_df = parse_weights(results = result, key = key, 
                                     include_as = include_as)
            df = pd.concat([df, cur_df.copy()])
            
    else:
        df = parse_weights(results = results, include_as = include_as)

    return df


def get_selection(weights_df: pd.DataFrame, 
                  order_groups: bool=False) -> pd.DataFrame:
    """
    This function takes a pd.DataFrame created by 
    `scmkl.get_weights()` and returns a selection table. Selection 
    refers to how many times a group had a nonzero group weight. To 
    calculate this, a col is added indicating whether the group was 
    selected. Then, the dataframe is grouped by alpha and group. 
    Selection can then be summed returning a dataframe with cols 
    `['Alpha', 'Group', Selection]`. If is the result of multiclass 
    run(s), `'Class'` column must be present and will be in resulting 
    df as well.

    Parameters
    ----------
    weights_df : pd.DataFrame
        A dataframe output by `scmkl.get_weights()` with cols
        `['Alpha', 'Group', 'Kernel Weight']`. If is the result of 
        multiclass run(s), `'Class'` column must be present as well.

    order_groups : bool
        If `True`, the `'Group'` col of the output dataframe will be 
        made into a `pd.Categorical` col ordered by number of times 
        each group was selected in decending order.

    Returns
    -------
    df : pd.DataFrame
        A dataframe with cols `['Alpha', 'Group', Selection]`. Also, 
        `'Class'` column if is a multiclass result.

    Example
    -------
    >>> # For a single file
    >>> results = scmkl.run(adata)
    >>> weights = scmkl.get_weights(results = results)
    >>> selection = scmkl.get_selection(weights)
    
    >>> # For multiple runs saved in a dict
    >>> output_dir = 'scMKL_outputs/'
    >>> rfiles = scmkl.read_files(output_dir)
    >>> weights = scmkl.get_weights(rfiles=rfiles)
    >>> selection = scmkl.get_selection(weights)
    """
    # Adding col indicating whether or not groups have nonzero weight
    selection = weights_df['Kernel Weight'].apply(lambda x: x > 0)
    weights_df['Selection'] = selection

    # Summing selection across replications to get selection
    is_mult = 'Class' in weights_df.columns
    if is_mult:
        df = weights_df.groupby(['Alpha', 'Group', 'Class'])['Selection'].sum()
    else:
        df = weights_df.groupby(['Alpha', 'Group'])['Selection'].sum()
    df = df.reset_index()

    # Getting group order
    if order_groups and not is_mult:
        order = df.groupby('Group')['Selection'].sum()
        order = order.reset_index().sort_values(by = 'Selection', 
                                                ascending = False)
        order = order['Group']
        df['Group'] = pd.Categorical(df['Group'], categories = order)


    return df


def groups_per_alpha(selection_df: pd.DataFrame) -> dict:
    """
    This function takes a pd.DataFrame from `scmkl.get_selection()` 
    generated from multiple scMKL results and returns a dictionary 
    with keys being alphas from the input dataframe and values being 
    the mean number of selected groups for a given alpha across 
    results. 

    Parameters
    ----------
    selection_df : pd.DataFrame
        A dataframe output by `scmkl.get_selection()` with cols 
        `['Alpha', 'Group', Selection].
    
    Returns
    -------
    mean_groups : dict
        A dictionary with alphas as keys and the mean number of 
        selected groups for that alpha as keys.

    Examples
    --------
    >>> weights = scmkl.get_weights(rfiles)
    >>> selection = scmkl.get_selection(weights)
    >>> mean_groups = scmkl.mean_groups_per_alpha(selection)
    >>> mean_groups = {alpha : np.round(num_selected, 1)
    ...                for alpha, num_selected in mean_groups.items()}
    >>>
    >>> print(mean_groups)
    {0.05 : 50.0, 0.2 : 24.7, 1.1 : 5.3}
    """
    mean_groups = {}
    for alpha in np.unique(selection_df['Alpha']):

        # Capturing rows for given alpha
        rows = selection_df['Alpha'] == alpha

        # Adding mean number of groups for alpha
        mean_groups[alpha] = np.mean(selection_df[rows]['Selection'])

    return mean_groups


def read_gtf(path: str, filter_to_coding: bool=False):
    """
    Reads and formats a gtf file. Adds colnames: `['chr', 'source', 
    'feature', 'start', 'end', 'score', 'strand', 'frame', 
    'attribute']`. If `filter_to_coding == True`, `'gene_name'` col 
    will be added with gene_name from attribute col if gene is protein 
    coding. If `'gene_name'` is not in attribute for that row, 
    `'gene_id'` will be used. 

    Parameters
    ----------
    path : str
        The file path to the gtf file to be read in. If the file is 
        gzipped, file name must end with .gz.

    filter_to_coding : bool
        If `True`, will filter rows in gtf data frame to only 
        protein coding genes. Will add column `'gene_name'` containing 
        the gene name for each row. If gene name is missing in GTF 
        gene_id will be used.

    Returns
    -------
    df : pd.DataFrame
        A pandas dataframe of the input gtf file.

    Examples
    --------
    >>> import scmkl
    >>>
    >>> file = 'data/hg38_subset_protein_coding.annotation.gtf'
    >>> gtf = scmkl.read_gtf(file)
    >>>
    >>> gtf.head()
            chr  source     feature  start    end score strand frame                                          
    0  chr1  HAVANA        gene  11869  14409     .      +     .  
    1  chr1  HAVANA  transcript  11869  14409     .      +     .  
    2  chr1  HAVANA        exon  11869  12227     .      +     .  
    3  chr1  HAVANA        exon  12613  12721     .      +     .  
    4  chr1  HAVANA        exon  13221  14409     .      +     .  
    attribute
    gene_id "ENSG00000223972.5"; gene_type "transc...
    gene_id "ENSG00000223972.5"; transcript_id "EN...
    gene_id "ENSG00000223972.5"; transcript_id "EN...
    gene_id "ENSG00000223972.5"; transcript_id "EN...
    gene_id "ENSG00000223972.5"; transcript_id "EN...
    """
    df = pd.read_csv(path, sep='\t', comment='#', 
                     skip_blank_lines=True, header=None)
    
    df.columns = ['chr', 'source', 'feature', 'start', 'end', 
                  'score', 'strand', 'frame', 'attribute']
    
    if filter_to_coding:
        prot_rows = df['attribute'].str.contains('protein_coding')
        df = df[prot_rows]
        df = df[df['feature'] == 'gene']

        # Capturing and adding gene name to df
        gene_names = list()

        for attr in df['attribute']:
            gname = re.findall(r'(?<=gene_name ")[A-z0-9\-]+', attr)

            if gname:
                gene_names.append(gname[0])
            else:
                gid = re.findall(r'(?<=gene_id ")[A-z0-9\-]+', attr)
                
                if gid:
                    gene_names.append(gid[0])
                else:
                    gene_names.append('NA')

        df['gene_name'] = gene_names
        df = df[df['gene_name'] != 'NA']
    
    return df


import pickle
import numpy as np
import pandas as pd
import gseapy as gp


# Prevent iteration of all availible gene sets for human and mouse
global_lib_orgs = ['human', 'mouse']

human_genesets = [
    'Azimuth_2023', 
    'Azimuth_Cell_Types_2021', 
    'Cancer_Cell_Line_Encyclopedia', 
    'CellMarker_2024', 
    'CellMarker_Augmented_2021', 
    'GO_Biological_Process_2025', 
    'GO_Cellular_Component_2025', 
    'GO_Molecular_Function_2025', 
    'KEGG_2019_Mouse', 
    'KEGG_2021_Human', 
    'MSigDB_Hallmark_2020', 
    'NCI-60_Cancer_Cell_Lines', 
    'WikiPathways_2024_Human', 
    'WikiPathways_2024_Mouse'
]


def check_organism(organism: str):
    """
    Makes sure that organism is in availible organisms from `gseapy`.

    Parameters
    ----------
    organism : str
        The organsim to check.

    Returns
    -------
    is_in : bool
        `True` if the organism is valid. `False` if organism is not an 
        option.
    """
    org_opts = {'human', 'mouse', 'yeast', 'fly', 'fish', 'worm'}
    org_err = f"Invalid `organism`, choose from {org_opts}"
    assert organism.lower() in org_opts, org_err

    return None


def check_groups(groups: list, key_terms: str | list='', 
                 blacklist: str | list | bool=False, other_org: str=''):
    """
    Takes a list of groups from a gene set library and checks the names 
    for the desired gene sets. Returns a dictionary with keys 
    `'key_terms_in'` that is `list` of `bool`s corresponding to 
    `names`. `'num_groups'` value is an int for how many groups are in 
    the library.

    Parameters
    ----------
    groups : list
        The names of groups for a given library.

    key_terms : str | list
        The types of cells or other specifiers the gene set is for 
        (example: 'CD4 T', 'kidney', ect...).

    other_org : str
        Either `'human'` or `'mouse'`. The organsim to ignore when 
        checking groups. Should be empty if target organism is not 
        `'human'` or `'mouse'`.

    Returns
    -------
    result : dict
        A dictionary with the following keys and values:

        `'name'` : list
            The names of each group in input `'groups'` if the name 
            does not contain `'other_org'`.
    
        `'key_terms_in'` : list
            A boolean list repective to `'name'` indicating if at 
            least one key word from `'key_terms'` is present in the group name.

        `'num_groups'` : int
            The length of `result['name']`.
    """
    result = {
        'key_terms_in' : list(),
        'blacklist_in' : list(),
        'name' : list(),
    }

    for group_name in groups:
            
        if not other_org in group_name.lower():

            if list == type(key_terms):
                key_terms_in = any([k.lower() in group_name.lower() 
                                    for k in key_terms])
            else:
                key_terms_in = key_terms.lower() in group_name.lower()

            if list == type(blacklist):
                blacklist_in = any([bl.lower() in group_name.lower() 
                                    for bl in blacklist])
            elif str == type(blacklist):
                blacklist_in = blacklist.lower() in group_name.lower()
            else:
                blacklist_in = False
                
            result['key_terms_in'].append(key_terms_in)
            result['blacklist_in'].append(blacklist_in)
            result['name'].append(group_name)

    result['num_groups'] = len(result['name'])

    return result


def check_libs(libs, key_terms: str | list='', 
               blacklist: str | list | bool=False, other_org: str=''):
    """
    Checks libraries for desired `key_terms` in groups.

    Parameters
    ----------
    libs : dict
        A dictionary as `libs[library_name] = library_groups`.

    key_terms : str | list
        A `str` or `list` of `str`s to seach for in `libs` group names.

    other_org : str
        Only applicable when desired organism is `'human'` or 
        `'mouse'`. If desired organism is `'human'`, `other_org` 
        should be `'mouse'` and vice-versa.

    Returns
    -------
    summary, tally : pd.DataFrame | pd.DataFrame
        `summary` has cols `['Library', 'No. Gene Sets', 
        'No. Key Terms Matching']` where `'Library'` is the library from 
        `gseapy` with `'No. Gene Sets'` and `'No. Key Terms Matching'` 
        corresponding. `'No. Key Terms Matching'` only included if 
        `key_terms` argument is provided. `tally` has cols `['library', 
        'key_terms_in', 'name']` 
    """
    num_groups = dict()
    
    tally = {
        'library' : list(), 
        'key_terms_in' : list(), 
        'blacklist_in' : list(),
        'name' : list()
    }
    
    for library, groups in libs.items():
        res = check_groups(list(groups.keys()), key_terms, 
                           blacklist, other_org)

        lib_repeats = [library]*len(res['name'])

        tally['library'].extend(lib_repeats)
        tally['key_terms_in'].extend(res['key_terms_in'])
        tally['blacklist_in'].extend(res['blacklist_in'])
        tally['name'].extend(res['name'])

        num_groups[library] = res['num_groups']

    tally = pd.DataFrame(tally)

    key_dict = tally.copy()
    key_dict = key_dict.groupby('library')['key_terms_in'].sum().reset_index()
    key_dict = dict(zip(key_dict['library'], key_dict['key_terms_in']))

    bl_dict = tally.copy()
    bl_dict = tally.groupby('library')['blacklist_in'].sum().reset_index()
    bl_dict = dict(zip(bl_dict['library'], bl_dict['blacklist_in']))

    lib_names = np.unique(tally['library'])
    key_counts = [key_dict[l] for l in lib_names]
    bl_counts = [bl_dict[l] for l in lib_names]
    n_groups = [num_groups[l] for l in lib_names]

    summary = {
        'Library' : lib_names,
        'No. Gene Sets' : n_groups
    }

    if key_terms:
        summary['No. Key Terms Matching'] = key_counts
    if blacklist:
        summary['No. Blacklist Matching'] = bl_counts

    summary = pd.DataFrame(summary)

    return summary, tally


def find_candidates(organism: str='human', key_terms: str | list='', blacklist: str | list | bool=False):
    """
    Given `organism` and `key_terms`, will search for gene 
    groupings that could fit the datasets/classification task. 
    `blacklist` terms undesired in group names.

    Parameters
    ----------
    organism : str
        The species the gene grouping is for. Options are 
        `{'Human', 'Mouse', 'Yeast', 'Fly', 'Fish', 'Worm'}`

    key_terms : str | list
        The types of cells or other specifiers the gene set is for 
        (example: 'CD4 T').

    blacklist : str | list | bool
        Term(s) undesired in group names. Ignored unless provided.

    Returns
    -------
    libraries : list
        A list of gene set library names that could serve for the 
        dataset/classification task.        

    Examples
    --------
    >>> scmkl.find_candidates('human', key_terms=' b ')
                                Library  No. Gene Sets
    0                    Azimuth_2023           1241
    1         Azimuth_Cell_Types_2021            341
    2   Cancer_Cell_Line_Encyclopedia            967
    3                 CellMarker_2024           1134
    No. Key Type Matching
    9
    9
    0
    21
    """
    check_organism(organism)

    if organism.lower() in global_lib_orgs:
        glo = global_lib_orgs.copy()
        glo.remove(organism)
        other_org = glo[0]
        libs = human_genesets
        libs = [name for name in libs if not other_org in name.lower()]
    else:
        libs = gp.get_library_name(organism)
        other_org = ''

    libs = {name : gp.get_library(name, organism)
            for name in libs}
    
    libs_df, _ = check_libs(libs, key_terms, blacklist, other_org)

    return libs_df


def get_gene_groupings(lib_name: str, organism: str='human', key_terms: str | list='', 
                       blacklist: str | list | bool=False, min_overlap: int=2,
                      genes: list | tuple | pd.Series | np.ndarray | set=[]):
    """
    Takes a gene set library name and filters to groups containing 
    element(s) in `key_terms`. If genes is provided, will 
    ensure that there are at least `min_overlap` number of genes in 
    each group. Resulting groups will meet all of the before-mentioned 
    criteria if `isin_logic` is `'and'` | `'or'`.

    Parameters
    ----------
    lib_name : str
        The desired library name.

    organism : str
        The species the gene grouping is for. Options are 
        `{'Human', 'Mouse', 'Yeast', 'Fly', 'Fish', 'Worm'}`.

    key_terms : str | list
        The types of cells or other specifiers the gene set is for 
        (example: 'CD4 T').

    genes : array_like
        A vector of genes from the reference/query datasets. If not 
        assigned, function will not filter groups based on feature 
        overlap.

    min_overlap : int
        The minimum number of genes that must be present in a group 
        for it to be kept. If `genes` is not given, ignored.

    Returns
    -------
    lib : dict
        The filtered library as a `dict` where keys are group names 
        and keys are features.

    Examples
    --------
    >>> dataset_feats = [
    ...    'FUCA1', 'CLIC4', 'STMN1', 'SYF2', 'TAS1R1', 
    ...    'NOL9', 'TAS1R3', 'SLC2A5', 'THAP3', 'IGHM', 
    ...    'MARCKS', 'BANK1', 'TNFRSF13B', 'IGKC', 'IGHD', 
    ...    'LINC01857', 'CD24', 'CD37', 'IGHD', 'RALGPS2'
    ...    ]
    >>> rna_grouping = scmkl.get_gene_groupings(
    ...   'Azimuth_2023', key_terms=[' b ', 'b cell', 'b '], 
    ...   genes=dataset_feats)
    >>>
    >>> rna_groupings.keys()
    dict_keys(['PBMC-L1-B Cell', 'PBMC-L2-Intermediate B Cell', ...])
    """
    check_organism(organism)
    
    lib = gp.get_library(lib_name, organism)

    if organism.lower() in global_lib_orgs:
        glo = global_lib_orgs.copy()
        glo.remove(organism)
        other_org = glo[0]
    else:
        other_org = ''

    group_names = list(lib.keys())
    res = check_groups(group_names, key_terms, blacklist, other_org)
    del res['num_groups']

    # Finding groups where group name matches key_terms
    g_summary = pd.DataFrame(res)

    if key_terms:
        kept = g_summary['key_terms_in']
        kept_groups = g_summary['name'][kept].to_numpy()
        g_summary = g_summary[kept]
    else:
        print("Not filtering with `key_terms` parameter.")
        kept_groups = g_summary['name'].to_numpy()

    if blacklist:
        kept = ~g_summary['blacklist_in']
        kept_groups = g_summary['name'][kept].to_numpy()
    else:
        print("Not filtering with `blacklist` parameter.")
    
    # Filtering library
    lib = {group : lib[group] for group in kept_groups}

    if 0 < len(genes):
        del_groups = list()
        genes = list(set(genes.copy()))
        for group, features in lib.items():
            overlap = np.isin(features, genes)
            overlap = np.sum(overlap)
            if overlap < min_overlap:
                print(overlap, flush=True)
                del_groups.append(group)

        # Removing genes without enough overlap
        for group in del_groups:
            print(f'Removing {group} from grouping.')
            del lib[group]

    else:
        print("Not checking overlap between group and dataset features.")

    return lib
